from __future__ import division, print_function

try:
    range = xrange
except NameError:
    pass

import os
import math
from pprint import pprint

ppr = pprint


class RSMDef(object):
    """
    % Rural Site & Vertical Diffusion Model (VDM)
    % Calculates the vertical profiles of air temperature above the weather
    % station per 'The uwg' (2012) Eq. 4, 5, 6.

    properties
        lat;           % latitude (deg)
        lon;           % longitude (deg)
        GMT;           % GMT hour correction
        height         % average obstacle height (m)
        z0r;           % rural roughness length (m)
        disp;          % rural displacement length (m)
        z;             % vertical height (m)
        dz;            % vertical discretization (m)
        nz0;           % layer number at zmt (m)
        nzref;         % layer number at zref (m)
        nzfor;         % layer number at zfor (m)
        nz10;          % layer number at zmu (m)
        nzi;           % layer number at zi_d (m)
        tempProf;      % potential temperature profile at the rural site (K)
        presProf;      % pressure profile at the rural site (Pa)
        tempRealProf;  % real temperature profile at the rural site (K)
        densityProfC;  % density profile at the center of layers (kg m-3)
        densityProfS;  % density profile at the sides of layers (kg m-3)
        windProf;      % wind profile at the rural site (m s-1)
        ublPres;       % Average pressure at UBL (Pa)
    end
    """

    Z_MESO_FILE_NAME = "z_meso.txt"

    def __init__(self,lat,lon,GMT,height,T_init,P_init,parameter,z_meso_path):

        # defines self.z_meso property
        self.load_z_meso(z_meso_path)

        self.lat = lat                  # latitude (deg)
        self.lon = lon                  # longitude (deg)
        self.GMT = GMT                  # GMT hour correction
        self.height = height            # average obstacle height (m)
        self.z0r = 0.1 * height         # rural roughness length (m)
        self.disp = 0.5 * height        # rural displacement lenght (m)

        # vertical grid at the rural site
        self.z  = [0 for x in range(len(self.z_meso)-1)] # Midht btwn each distance interval
        self.dz = [0 for x in range(len(self.z_meso)-1)] # Distance betweeen each interval

        for zi in range(len(self.z_meso)-1):
            self.z[zi] = 0.5 * (self.z_meso[zi] + self.z_meso[zi+1])
            self.dz[zi] = self.z_meso[zi+1] - self.z_meso[zi]

        # Define initial booleans
        ll = True
        mm = True
        nn = True
        oo = True
        pp = True

        # Define self.nz0, self.nzref, self.nzfor, self.nz10, self.nzi
        for iz in range(len(self.z_meso)-1):
            # self.nz0: self.z index >= reference height for weather station
            eq_th = self.is_near_zero(self.z[iz] - parameter.tempHeight)
            if (eq_th == True or self.z[iz] > parameter.tempHeight) and ll==True:
                self.nz0 = iz+1   # layer number at zmt (m)
                ll = False

            # self.nzref: self.z index >= reference inversion height
            eq_rh = self.is_near_zero(self.z[iz] - parameter.refHeight)
            if (eq_rh == True or self.z[iz] > parameter.refHeight) and mm==True:
              self.nzref = iz+1   # layer number at zref (m)
              mm = False

            # self.nzfor: self.z index >= nighttime boundary layer height
            eq_nh = self.is_near_zero(self.z[iz] - parameter.nightBLHeight)
            if (eq_nh == True or self.z[iz] > parameter.nightBLHeight) and nn==True:
              self.nzfor = iz+1   # layer number at zfor (m)
              nn = False

            # self.nz10: self.z index >= wind height
            eq_wh = self.is_near_zero(self.z[iz] - parameter.windHeight)
            if (eq_wh == True or self.z[iz] > parameter.windHeight) and oo==True:
              self.nz10 = iz+1    # layer number at zmu (m)
              oo = False

            eq_dh = self.is_near_zero(self.z[iz] - parameter.dayBLHeight)
            if (eq_dh == True or self.z[iz] > parameter.dayBLHeight) and pp==True:
              self.nzi = iz+1     # layer number at zi_d (m)
              pp = False

        # Define temperature, pressure and density vertical profiles
        self.tempProf = [T_init for x in range(self.nzref)]
        self.presProf = [P_init for x in range(self.nzref)]

        for iz in range(1,self.nzref):
            self.presProf[iz] = (self.presProf[iz-1]**(parameter.r/parameter.cp) -\
               parameter.g/parameter.cp * (P_init**(parameter.r/parameter.cp)) * (1./self.tempProf[iz] +\
               1./self.tempProf[iz-1]) * 0.5 * self.dz[iz])**(1./(parameter.r/parameter.cp))

        self.tempRealProf = [T_init for x in range(self.nzref)]
        for iz in range(self.nzref):
           self.tempRealProf[iz] = self.tempProf[iz] * (self.presProf[iz] / P_init)**(parameter.r/parameter.cp)

        self.densityProfC = [None for x in range(self.nzref)]
        for iz in range(self.nzref):
           self.densityProfC[iz] = self.presProf[iz] / parameter.r / self.tempRealProf[iz]

        self.densityProfS = [self.densityProfC[0] for x in range(self.nzref+1)]
        for iz in range(1,self.nzref):
           self.densityProfS[iz] = (self.densityProfC[iz] * self.dz[iz-1] +\
               self.densityProfC[iz-1] * self.dz[iz]) / (self.dz[iz-1]+self.dz[iz])

        self.densityProfS[self.nzref] = self.densityProfC[self.nzref-1]
        self.windProf = [1 for x in range(self.nzref)]

    def __repr__(self):
        return "RSM: obstacle ht = {}m, surface roughness length = {}m, displacement length = {}m".format(
            self.height,
            self.z0r,         # rural roughness length (m)
            self.disp         # rural displacement lenght (m)
            )

    def is_near_zero(self,num,eps=1e-16):
        return abs(float(num)) < eps

    def load_z_meso(self,z_meso_path):
        """ Open the z_meso.txt file and return heights as list """

        self.z_meso = []
        z_meso_file_path = os.path.join(z_meso_path, self.Z_MESO_FILE_NAME)

        # Check if exists
        if not os.path.exists(z_meso_file_path):
            raise Exception("z_meso.txt file: '{}' does not exist.".format(uwg_param_file))

        f = open(z_meso_file_path,'r')
        for txtline in f:
            z_ = float("".join(txtline.split())) # Strip all white spaces and change to float
            self.z_meso.append(z_)
        f.close()


    # Ref: The uwg (2012), Eq. (4)
    def VDM(self,forc,rural,parameter,simTime):

        self.tempProf[0] = forc.temp    # Lower boundary condition

        # compute pressure profile
        for iz in reversed(list(range(self.nzref))[1:]):
           self.presProf[iz-1] = (math.pow(self.presProf[iz],parameter.r/parameter.cp) + \
               parameter.g/parameter.cp*(math.pow(forc.pres,parameter.r/parameter.cp)) * \
               (1./self.tempProf[iz] + 1./self.tempProf[iz-1]) * \
               0.5 * self.dz[iz])**(1./(parameter.r/parameter.cp))

        # compute the real temperature profile
        for iz in range(self.nzref):
            self.tempRealProf[iz]= self.tempProf[iz] * \
            (self.presProf[iz]/forc.pres)**(parameter.r/parameter.cp)

        # compute the density profile
        for iz in range(self.nzref):
           self.densityProfC[iz] = self.presProf[iz]/parameter.r/self.tempRealProf[iz]
        self.densityProfS[0] = self.densityProfC[0]

        for iz in range(1,self.nzref):
           self.densityProfS[iz] = (self.densityProfC[iz] * self.dz[iz-1] + \
               self.densityProfC[iz-1] * self.dz[iz])/(self.dz[iz-1] + self.dz[iz])

        self.densityProfS[self.nzref] = self.densityProfC[self.nzref-1]

        # Ref: The uwg (2012), Eq. (5)
        # compute diffusion coefficient
        cd,ustarRur = self.DiffusionCoefficient(self.densityProfC[0], \
            self.z, self.dz, self.z0r, self.disp, \
            self.tempProf[0], rural.sens, self.nzref, forc.wind, \
            self.tempProf, parameter)


        # solve diffusion equation
        self.tempProf = self.DiffusionEquation(self.nzref,simTime.dt,\
            self.tempProf,self.densityProfC,self.densityProfS,cd,self.dz)

        # compute wind profile
        # N.B In Matlab, negative values are converted to complex values.
        # log(-x) = log(x) + log(-1) = log(x) + i*pi
        # Python will throw an exception. Negative value occurs here if
        # VDM is run for average obstacle height ~ 4m.
        for iz in range(self.nzref):
            self.windProf[iz] = ustarRur/parameter.vk*\
                math.log((self.z[iz]-self.disp)/self.z0r)

        # Average pressure
        self.ublPres = 0.
        for iz in range(self.nzfor):
            self.ublPres = self.ublPres + \
                self.presProf[iz]*self.dz[iz]/(self.z[self.nzref-1]+self.dz[self.nzref-1]/2.)

    def DiffusionEquation(self,nz,dt,co,da,daz,cd,dz):

        cddz = [0 for i in range(nz+2)]
        a = [[0 for j in range(3)] for i in range(nz)]
        c = [0 for i in range(nz)]

        #--------------------------------------------------------------------------
        cddz[0] = daz[0]*cd[0]/dz[0]
        for iz in range(1,nz):
            cddz[iz] = 2.*daz[iz]*cd[iz]/(dz[iz]+dz[iz-1])
        cddz[nz] = daz[nz]*cd[nz]/dz[nz]
        #--------------------------------------------------------------------------
        a[0][0] = 0.
        a[0][1] = 1.
        a[0][2] = 0.
        c[0] = co[0]

        for iz in range(1,nz-1):
            dzv = dz[iz]
            a[iz][0]=-cddz[iz]*dt/dzv/da[iz]
            a[iz][1]=1+dt*(cddz[iz]+cddz[iz+1])/dzv/da[iz]
            a[iz][2]=-cddz[iz+1]*dt/dzv/da[iz]
            c[iz]=co[iz]

        a[nz-1][0]=-1.
        a[nz-1][1]=1.
        a[nz-1][2]=0.
        c[nz-1]=0.

        #--------------------------------------------------------------------------
        co = self.invert(nz,a,c)
        return co

    def DiffusionCoefficient(self,rho,z,dz,z0,disp,tempRur,heatRur,nz,uref,th,parameter):
        # Initialization
        Kt = [0 for x in range(nz+1)]
        ws = [0 for x in range(nz)]
        te = [0 for x in range(nz)]
        # Friction velocity (Louis 1979)
        ustar = parameter.vk * uref/math.log((10.-disp)/z0)

        # Monin-Obukhov length
        lengthRur = max(-rho*parameter.cp*ustar**3*tempRur/parameter.vk/parameter.g/heatRur,-50.)

        # Unstable conditions
        if heatRur > 1e-2:
            # Convective velocity scale
            wstar = (parameter.g*heatRur*parameter.dayBLHeight/rho/parameter.cp/tempRur)**(1/3.)
            # Wind profile function
            phi_m = (1-8.*0.1*parameter.dayBLHeight/lengthRur)**(-1./3.)

            for iz in range(nz):
                # Mixed-layer velocity scale
                ws[iz] = (ustar**3 + phi_m*parameter.vk*wstar**3*z[iz]/parameter.dayBLHeight)**(1/3.)
                # TKE approximation
                te[iz] = max(ws[iz]**2., 0.01)

        else: # Stable and neutral conditions
            for iz in range(nz):
                # TKE approximation
                te[iz] = max(ustar**2.,0.01)

        # lenght scales (l_up, l_down, l_k, l_eps)
        self.dlu, self.dld = self.DissipationBougeault(parameter.g,nz,z,dz,te,th)

        self.dld,dls,dlk = self.LengthBougeault(nz,self.dld,self.dlu,z)

        # Boundary-layer diffusion coefficient
        for iz in range(nz):
           Kt[iz] = 0.4*dlk[iz]*math.sqrt(te[iz])

        Kt[nz] = Kt[nz-1]

        return Kt, ustar

    def DissipationBougeault(self,g,nz,z,dz,te,pt):
        # Note on translation from UWG_Matlab
        # list length (i.e nz) != list indexing (i.e dlu[0] in python
        # wherease in matlab it is

        dlu = [0 for x in range(nz)]
        dld = [0 for x in range(nz)]

        for iz in range(nz):
            zup=0.
            dlu[iz] = z[nz] - z[iz] - dz[iz]/2.
            zzz=0.
            zup_inf=0.
            beta=g/pt[iz]

            for izz in range(iz,nz-1):
                dzt=(dz[izz+1]+dz[izz])/2.
                zup=zup-beta*pt[iz]*dzt
                zup=zup+beta*(pt[izz+1]+pt[izz])*dzt/2.
                zzz=zzz+dzt

                if (te[iz]<zup) and ((te[iz]>zup_inf) or self.is_near_zero(te[iz]-zup_inf)):
                    bbb=(pt[izz+1]-pt[izz])/dzt

                    if not self.is_near_zero(bbb-0.):
                        tl=(-beta*(pt[izz]-pt[iz])+ \
                        math.sqrt( max(0.,(beta*(pt[izz]-pt[iz]))**2.+ \
                        2.*bbb*beta*(te[iz]-zup_inf))))/bbb/beta
                    else:
                        tl=(te[iz]-zup_inf)/(beta*(pt[izz]-pt[iz]))
                    dlu[iz]=max(1.,zzz-dzt+tl)
                zup_inf=zup

            zdo=0.
            zdo_sup=0.
            dld[iz]=z[iz]+dz[iz]/2.
            zzz=0.

            for izz in range(iz,0,-1):
                dzt=(dz[izz-1]+dz[izz])/2.
                zdo=zdo+beta*pt[iz]*dzt
                zdo=zdo-beta*(pt[izz-1]+pt[izz])*dzt/2.
                zzz=zzz+dzt

                if (te[iz]<zdo) and ((te[iz]>zdo_sup) or self.is_near_zero(te[iz]-zdo_sup)):
                    bbb=(pt[izz]-pt[izz-1])/dzt

                    if not self.is_near_zero(bbb-0.):
                        tl=(beta*(pt[izz]-pt[iz])+ \
                            math.sqrt( max(0.,(beta*(pt[izz]-pt[iz]))**2.+ \
                            2.*bbb*beta*(te[iz]-zdo_sup))))/bbb/beta
                    else:
                        tl=(te[iz]-zdo_sup)/(beta*(pt[izz]-pt[iz]))
                    dld[iz]=max(1.,zzz-dzt+tl)
                zdo_sup=zdo

        return dlu,dld

    def LengthBougeault(self,nz,dld,dlu,z):

        dlg = [0 for x in range(nz)]
        dls = [0 for x in range(nz)]
        dlk = [0 for x in range(nz)]

        for iz in range(nz):
            dlg[iz] = (z[iz]+z[iz+1])/2.

        for iz in range(nz):
            dld[iz] = min(dld[iz], dlg[iz])
            dls[iz] = math.sqrt(dlu[iz]*dld[iz])
            dlk[iz] = min(dlu[iz],dld[iz])

        return dld,dls,dlk

    def invert(self,nz,A,C):
        """
        Inversion and resolution of a tridiagonal matrix
                 A X = C
        Input:
         nz number of layers
         a(*,1) lower diagonal (Ai,i-1)
         a(*,2) principal diagonal (Ai,i)
         a(*,3) upper diagonal (Ai,i+1)
         c
        Output
         x     results
        """

        X = [0 for i in range(nz)]

        for i in reversed(range(nz-1)):
            C[i] = C[i] - A[i][2] * C[i+1]/A[i+1][1]
            A[i][1] = A[i][1] - A[i][2] * A[i+1][0]/A[i+1][1]

        for i in  range(1,nz,1):
            C[i] = C[i] - A[i][0] * C[i-1]/A[i-1][1]

        for i in range(nz):
            X[i] = C[i]/A[i][1]

        return X



class Param(object):
    """
    Geographic Parameters
    """

    def __init__(self, dayBLHeight, nightBLHeight, refHeight, tempHeight, windHeight,
                circCoeff, dayThreshold, nightThreshold, treeFLat, grassFLat, vegAlbedo, vegStart,
                vegEnd, nightSetStart, nightSetEnd, windMin, wgmax, exCoeff, maxdx, g,
                cp, vk, r, rv, lv, pi, sigma, waterDens, lvtt, tt, estt, cl, cpv, b, cm, colburn):
        self.dayBLHeight = dayBLHeight           # daytime mixing height, orig = 700
        self.nightBLHeight = nightBLHeight       # Sing: 80, Bub-Cap: 50, nighttime boundary-layer height (m); orig 80
        self.refHeight = refHeight               # Reference height at which the vertical profile of potential temperature is vertical
        self.tempHeight = tempHeight             # Temperature measuremnt height at the weather station (m)
        self.windHeight = windHeight             # Air velocity measuremnt height at the weather station (m)
        self.circCoeff =  circCoeff              # Wind scaling coefficient
        self.dayThreshold = dayThreshold         # heat flux threshold for daytime conditions (W m-2)
        self.nightThreshold = nightThreshold     # heat flux threshold for nighttime conditions (W m-2)
        self.treeFLat = treeFLat                 # latent fraction of trees
        self.grassFLat = grassFLat               # latent fraction of grass
        self.vegAlbedo = vegAlbedo               # albedo of vegetation
        self.vegStart = vegStart                 # begin month for vegetation participation
        self.vegEnd = vegEnd                     # end month for vegetation participation
        self.nightSetStart = nightSetStart       # begin hour for night thermal set point schedule
        self.nightSetEnd = nightSetEnd           # end hour for night thermal set point schedule
        self.windMin = windMin                   # minimum wind speed (m s-1)
        self.wgmax = wgmax                       # maximum film water depth on horizontal surfaces (m)
        self.exCoeff = exCoeff                   # exchange velocity coefficient
        self.maxdx = maxdx                       # maximum discretization length for the UBL model (m)
        self.g = g                               # gravity (m s-2)
        self.cp = cp                             # heat capacity for air (J/kg K)
        self.vk = vk                             # von karman constant (dimensionless)
        self.r = r                               # gas constant for dry air (J/kg K)
        self.rv = rv                             # gas constant for water vapor (J/kg K)
        self.lv = lv                             # latent heat of evaporation (J/kg)
        self.pi = pi                             # pi
        self.sigma = sigma                       # Stefan Boltzmann constant (W K-4 m-2)
        self.waterDens = waterDens               # water density (kg m-3)
        self.lvtt = lvtt                         # ?
        self.tt = tt                             # ?
        self.estt = estt                         # ?
        self.cl = cl                             # ?
        self.cpv = cpv                           # ?
        self.b   = b                             # coefficients derived by Louis (1979)
        self.cm  = cm                            # ?
        self.colburn = colburn                   # (Pr/Sc)^(2/3) for Colburn analogy in water evaporation

    def __repr__(self):
        b=self.treeFLat
        return "geoParameter: dayBLht = {}m, nightBLht = {}m, station ht = {}m".format(
            self.dayBLHeight,
            self.nightBLHeight,
            self.tempHeight
            )



from __future__ import division

from math import log, pow, exp


def psychrometrics (Tdb_in, w_in, P):
    """
    Modified version of Psychometrics by Tea Zakula
    MIT Building Technology Lab
    Input: Tdb_in, w_in, P
    Output: Tdb, w, phi, h, Tdp, v

    where:
    Tdb_in = [K] dry bulb temperature
    w_in = [kgv/kgda] Humidity Ratio
    P = [P] Atmospheric Station Pressure

    Tdb: [C] dry bulb temperature
    w: [kgv/kgda] Humidity Ratio
    phi: [Pw/Pws*100] relative humidity
    Tdp: [C] dew point temperature
    h: [J/kga] enthalpy
    v: [m3/kga] specific volume
    """
    # Change units
    c_air = 1006.   # [J/kg] air heat capacity, value from ASHRAE Fundamentals
    hlg = 2501000.  # [J/kg] latent heat, value from ASHRAE Fundamentals
    cw  = 1860.     # [J/kg] value from ASHRAE Fundamentals
    P = P/1000.     # convert from Pa to kPa

    Tdb = Tdb_in - 273.15
    w = w_in

    # phi (RH) calculation from Tdb and w
    Pw = (w*P)/(0.621945 + w)                             # partial pressure of water vapor
    Pws = saturation_pressure(Tdb)                        # Get saturation pressure for given Tdb
    phi = Pw/Pws*100.0

    # enthalpy calculation from Tdb and w
    h = c_air*Tdb + w*(hlg+cw*Tdb)                      # [J kga-1]

    # specific volume calculation from Tdb and w
    v = 0.287042 * (Tdb+273.15)*(1+1.607858*w)/P        # ?

    # dew point calculation from w
    _pw = (w*P)/(0.621945 + w) # water vapor partial pressure in kPa
    alpha = log(_pw)

    Tdp = 6.54 + 14.526*alpha + pow(alpha,2)*0.7389 + pow(alpha,3)*0.09486 + pow(_pw,0.1984)*0.4569  # valid for Tdp between 0 C and 93 C

    return  Tdb, w, phi, h, Tdp, v

def saturation_pressure(Tdb_):
    T = Tdb_ + 273.15

    # N.B In Matlab, negative values are converted to complex values.
    # log(-x) = log(x) + log(-1) = log(x) + i*pi
    # Python will throw an exception. Negative value occurs here if
    # simulation timestep (dtSim) is large, i.e 3600s.
    _Pws = exp(-1*(5.8002206e3) / T+1.3914993 + (4.8640239e-2)*T*(-1.) + (4.1764768e-5)*pow(T,2) - (1.4452093e-8)*pow(T,3) + 6.5459673*log(T))  #in Pa
    _Pws = _Pws/1000.                                                               # in kPa
    return _Pws

def moist_air_density(P,Tdb,H):
    # Moist air density [kgv/ m-3] given dry bulb temperature, humidity ratio, and pressure.
    # ASHRAE Fundamentals (2005) ch. 6 eqn. 28
    # ASHRAE Fundamentals (2009) ch. 1 eqn. 28
    # from: https://github.com/psychrometrics/Libraries/blob/master/Psychrometrics_SI.cpp
    moist_air_density = P/(1000*0.287042*Tdb*(1.+1.607858*H))
    return moist_air_density

def HumFromRHumTemp(RH,T,P):
    # Derive Specific HUmidity [kgh20/kgn202] from RH, T and Pa
    # Saturation vapour pressure from ASHRAE
    C8 = -5.8002206e3
    C9 = 1.3914993
    C10 = -4.8640239e-2
    C11 = 4.1764768e-5
    C12 = -1.4452093e-8
    C13 = 6.5459673

    T += 273.15

    PWS = exp(C8/T + C9 + C10*T + C11 * pow(T,2) + C12 * pow(T,3) + C13 * log(T))
    PW = RH*PWS/100.0        # Vapour pressure
    W = 0.62198*PW/(P-PW)    # 4. Specific humidity
    return W

"""
function psat = psat(temp,parameter)
    gamw  = (parameter.cl - parameter.cpv) / parameter.rv;
    betaw = (parameter.lvtt/parameter.rv) + (gamw * parameter.tt);
    alpw = log(parameter.estt) + (betaw /parameter.tt) + (gamw *log(parameter.tt));
    psat = zeros(size(temp));
    for jj=1:size(temp)
        psat = exp(alpw - betaw/temp - gamw*log(temp));
    end
end

% Not used for this release but saved for possible future use
function Twb = wet_bulb(Tdb,Tdp,pres)

    % Copyright (c) 2015, Rolf Henry Goodwin
    % All rights reserved.
    %
    % Redistribution and use in source and binary forms, with or without
    % modification, are permitted provided that the following conditions are
    % met:
    %
    %     * Redistributions of source code must retain the above copyright
    %       notice, this list of conditions and the following disclaimer.
    %     * Redistributions in binary form must reproduce the above copyright
    %       notice, this list of conditions and the following disclaimer in
    %       the documentation and/or other materials provided with the distribution
    %
    % THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
    % AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
    % IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
    % ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
    % LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
    % CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
    % SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
    % INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
    % CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
    % ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
    % POSSIBILITY OF SUCH DAMAGE.

    % Code modified to merge into a single file - Joseph Yang, 2016


    % Tdb, Tdp, Twb in K
    % p in Pa (obtained function uses hPa, so /100 needed)
    global T;
    global T_d;
    global p;
    T = Tdb;
    T_d = Tdp;
    p = pres/100;

    Twb = root_finder(@Delta_q,T_d,T);
end

function dQTw = Delta_q(T_w)
    %Delta_q finds the value of function dq(Tw)
    %INPUT wet bulb temperature T_w
    %OUTPUT dq(Tw)
    global T;
    global T_d;
    global p;

    Cp = 1005; % Heat capacity of water vapor in J/(kg*K)
    L = 2.501e6; % Latent heat of water vapor at 0 degC in J/kg
    w1 = mixing_ratio(T_d,p); % Mixing ratio corresponding to T_d and p
    w2 = mixing_ratio(T_w,p); % Mixing ratio corresponding to T_w and p

    dQTw = (L*(w2-w1))/(1+w2)-Cp*(T-T_w)*(1+0.8*w2); % Finds deltaq(Tw)

end
"""



class SchDef(object):
    """
    Schedule class\

    Elec        % Schedule for electricity (WD,Sat,Sun)
    Gas         % Schedule for gas (WD,Sat,Sun)
    Light       % Schedule for light (WD,Sat,Sun)
    Occ         % Schedule for occupant (WD,Sat,Sun)
    Cool        % Temperature schedule for cooling (WD,Sat,Sun)
    Heat        % Temperature schedule for heating (WD,Sat,Sun)
    SWH         % Hot water schedule (tbd)

    % Internal Heat Load from DOE
    Qelec       % W/m^2 (max) for electrical plug process
    Qgas        % W/m^2 (max) for gas process
    Qlight      % W/m^2 (max) for light process
    Nocc        % #/m^2 (max) for occupancy
    Vswh        % Hot water vol/hr (max)
    Vent        % litres/s/person for ventilation
    """

    def __init__(self,Elec=None,Gas=None,Light=None,Occ=None,Cool=None,Heat=None,SWH=None):
        self.Elec = Elec
        self.Gas = Gas
        self.Light = Light
        self.Occ = Occ
        self.Cool = Cool
        self.Heat = Heat
        self.SWH = SWH

    def __repr__(self):
        return "Schedule:(weekday from 8:00 - 18:00) \n Heating: {a}\n Cooling: {b}".format(
            a= "Null" if self.Heat is None else self.Heat[0][7:17],
            b= "Null" if self.Cool is None else self.Cool[0][7:17]
            )



from .utilities import read_csv, str2fl
from math import pow, log, exp
from .psychrometrics import HumFromRHumTemp

try:
    range = xrange
except NameError:
    pass


class Weather(object):
    """
    Weather
    Read epw file
    http://bigladdersoftware.com/epx/docs/8-2/auxiliary-programs/epw-csv-format-inout.html
    properties
        location  # location name
        staTemp   % air temperature (C)
        staTdp    % dewpoint temperature (C)
        staRhum   % air relative humidity (%)
        staPres   % air pressure (Pa)
        staInfra  % horizontal Infrared Radiation Intensity (W m-2)
        staHor    % horizontal radiation
        staDir    % normal solar direct radiation (W m-2)
        staDif    % horizontal solar diffuse radiation (W m-2)
        staUdir   % wind direction ()
        staUmod   % wind speed (m s-1)
        staRobs   % Precipitation (mm h-1)
        staHum    % specific humidty (kg kg-1)
    """

    def __init__(self,climate_file,HI,HF):
        #HI: Julian start date
        #HF: Julian final date
        #H1 and HF define the row we want

        # Open .epw file and feed csv data to self.climate_data
        try:
            self.climate_data = read_csv(climate_file)
        except Exception as e:
            raise Exception("Failed to read .epw file! {}".format(e.message))

        self.location = self.climate_data[0][1]
        cd = self.climate_data[HI:HF+1]
        self.staTemp = str2fl([cd[i][6] for i in range(len(cd))])           # drybulb [C]
        self.staTdp = str2fl([cd[i][7] for i in range(len(cd))])            # dewpoint [C]
        self.staRhum = str2fl([cd[i][8] for i in range(len(cd))])           # air relative humidity (%)
        self.staPres = str2fl([cd[i][9] for i in range(len(cd))])           # air pressure (Pa)
        self.staInfra = str2fl([cd[i][12] for i in range(len(cd))])         # horizontal Infrared Radiation Intensity (W m-2)
        self.staHor = str2fl([cd[i][13] for i in range(len(cd))])           # horizontal radiation [W m-2]
        self.staDir = str2fl([cd[i][14] for i in range(len(cd))])           # normal solar direct radiation (W m-2)
        self.staDif = str2fl([cd[i][15] for i in range(len(cd))])           # horizontal solar diffuse radiation (W m-2)
        self.staUdir = str2fl([cd[i][20] for i in range(len(cd))])          # wind direction ()
        self.staUmod = str2fl([cd[i][21] for i in range(len(cd))])          # wind speed (m s-1)
        self.staRobs = str2fl([cd[i][33] for i in range(len(cd))])          # Precipitation (mm h-1)
        self.staHum = [0.0] * len(self.staTemp)                                     # specific humidty (kgH20 kgN202-1)
        for i in range(len(self.staTemp)):
            self.staHum[i] = HumFromRHumTemp(self.staRhum[i], self.staTemp[i], self.staPres[i])

        self.staTemp = [s+273.15 for s in self.staTemp]                             # air temperature (K)

    def __repr__(self):
        return "Weather: City = {}, Max Tdb = {}C, Min Tdb = {}C".format(
            self.location,
            max(self.staTemp)-273.15,
            min(self.staTemp)-273.15
            )



from __future__ import division, print_function

try:
    range = xrange
except NameError:
    pass

import sys
import os

try:
    import cPickle as pickle
except ImportError:
    import pickle

from .building import Building
from .material import Material
from .element import Element
from .BEMDef import BEMDef
from .schdef import SchDef
from .utilities import read_csv, str2fl

# For debugging only
#import pprint
#import decimal
#pp = pprint.pprint
#dd = decimal.Decimal.from_float

DIR_CURR = os.path.abspath(os.path.dirname(__file__))
DIR_DOE_PATH = os.path.join(DIR_CURR,"..","resources","DOERefBuildings")

# Define standards: 16 buiding types, 3 built eras, 16 climate zones

# DOE Building Types
BLDTYPE = [
    'FullServiceRestaurant',    # 1
    'Hospital',                 # 2
    'LargeHotel',               # 3
    'LargeOffice',              # 4
    'MedOffice',                # 5
    'MidRiseApartment',         # 6
    'OutPatient',               # 7
    'PrimarySchool',            # 8
    'QuickServiceRestaurant',   # 9
    'SecondarySchool',          # 10
    'SmallHotel',               # 11
    'SmallOffice',              # 12
    'StandAloneRetail',         # 13
    'StripMall',                # 14
    'SuperMarket',              # 15
    'WareHouse']                # 16

BUILTERA = [
    'Pre80',                    # 1
    'Pst80',                    # 2
    'New'                       # 3
    ]

ZONETYPE = [
    '1A (Miami)',               # 1
    '2A (Houston)',             # 2
    '2B (Phoenix)',             # 3
    '3A (Atlanta)',             # 4
    '3B-CA (Los Angeles)',      # 5
    '3B (Las Vegas)',           # 6
    '3C (San Francisco)',       # 7
    '4A (Baltimore)',           # 8
    '4B (Albuquerque)',         # 9
    '4C (Seattle)',             # 10
    '5A (Chicago)',             # 11
    '5B (Boulder)',             # 12
    '6A (Minneapolis)',         # 13
    '6B (Helena)',              # 14
    '7 (Duluth)',               # 15
    '8 (Fairbanks)'             # 16
    ]


def readDOE(serialize_output=True):
    """
    Read csv files of DOE buildings
    Sheet 1 = BuildingSummary
    Sheet 2 = ZoneSummary
    Sheet 3 = LocationSummary
    Sheet 4 = Schedules
    Note BLD8 & 10 = school


    Then make matrix of ref data as nested nested lists [16, 3, 16]:
    matrix refDOE = Building objs
    matrix Schedule = SchDef objs
    matrix refBEM (16,3,16) = BEMDef
    where:
        [16,3,16] is Type = 1-16, Era = 1-3, climate zone = 1-16
        i.e.
        Type: FullServiceRestaurant, Era: Pre80, Zone: 6A Minneapolis
    Nested tree:
    [TYPE_1:
        ERA_1:
            CLIMATE_ZONE_1
            ...
            CLIMATE_ZONE_16
        ERA_2:
            CLIMATE_ZONE_1
            ...
            CLIMATE_ZONE_16
        ...
        ERA_3:
            CLIMATE_ZONE_1
            ...
            CLIMATE_ZONE_16]

    """

    #Nested, nested lists of Building, SchDef, BEMDef objects
    refDOE   = [[[None]*16 for k_ in range(3)] for j_ in range(16)]                #refDOE(16,3,16) = Building
    Schedule = [[[None]*16 for k_ in range(3)] for j_ in range(16)]                #Schedule (16,3,16) = SchDef
    refBEM   = [[[None]*16 for k_ in range(3)] for j_ in range(16)]                #refBEM (16,3,16) = BEMDef

    #Purpose: Loop through every DOE reference csv and extract building data
    #Nested loop = 16 types, 3 era, 16 zones = time complexity O(n*m*k) = 768
    for i in range(16):

        #i = 16 types of buildings
        #print "\tType: {} @i={}".format(BLDTYPE[i], i)

        # Read building summary (Sheet 1)
        file_doe_name_bld = os.path.join("{}".format(DIR_DOE_PATH), "BLD{}".format(i+1),"BLD{}_BuildingSummary.csv".format(i+1))
        list_doe1 = read_csv(file_doe_name_bld)
        #listof(listof 3 era values)
        nFloor      = str2fl(list_doe1[3][3:6])      # Number of Floors, this will be list of floats and str if "basement"
        glazing     = str2fl(list_doe1[4][3:6])      # [?] Total
        hCeiling    = str2fl(list_doe1[5][3:6])      # [m] Ceiling height
        ver2hor     = str2fl(list_doe1[7][3:6])      # Wall to Skin Ratio
        AreaRoof    = str2fl(list_doe1[8][3:6])      # [m2] Gross Dimensions - Total area

        # Read zone summary (Sheet 2)
        file_doe_name_zone = os.path.join("{}".format(DIR_DOE_PATH), "BLD{}".format(i+1),"BLD{}_ZoneSummary.csv".format(i+1))
        list_doe2 = read_csv(file_doe_name_zone)
        #listof(listof 3 eras)
        AreaFloor   = str2fl([list_doe2[2][5],list_doe2[3][5],list_doe2[4][5]])       # [m2]
        Volume      = str2fl([list_doe2[2][6],list_doe2[3][6],list_doe2[4][6]])       # [m3]
        AreaWall    = str2fl([list_doe2[2][8],list_doe2[3][8],list_doe2[4][8]])       # [m2]
        AreaWindow  = str2fl([list_doe2[2][9],list_doe2[3][9],list_doe2[4][9]])       # [m2]
        Occupant    = str2fl([list_doe2[2][11],list_doe2[3][11],list_doe2[4][11]])    # Number of People
        Light       = str2fl([list_doe2[2][12],list_doe2[3][12],list_doe2[4][12]])    # [W/m2]
        Elec        = str2fl([list_doe2[2][13],list_doe2[3][13],list_doe2[4][13]])    # [W/m2] Electric Plug and Process
        Gas         = str2fl([list_doe2[2][14],list_doe2[3][14],list_doe2[4][14]])    # [W/m2] Gas Plug and Process
        SHW         = str2fl([list_doe2[2][15],list_doe2[3][15],list_doe2[4][15]])    # [Litres/hr] Peak Service Hot Water
        Vent        = str2fl([list_doe2[2][17],list_doe2[3][17],list_doe2[4][17]])    # [L/s/m2] Ventilation
        Infil       = str2fl([list_doe2[2][20],list_doe2[3][20],list_doe2[4][20]])    # Air Changes Per Hour (ACH) Infiltration

        # Read location summary (Sheet 3)
        file_doe_name_location = os.path.join("{}".format(DIR_DOE_PATH), "BLD{}".format(i+1),"BLD{}_LocationSummary.csv".format(i+1))
        list_doe3 = read_csv(file_doe_name_location)
        #(listof (listof 3 eras (listof 16 climate types)))
        TypeWall    = [list_doe3[3][4:20],list_doe3[14][4:20],list_doe3[25][4:20]]            # Construction type
        RvalWall    = str2fl([list_doe3[4][4:20],list_doe3[15][4:20],list_doe3[26][4:20]])     # [m2*K/W] R-value
        TypeRoof    = [list_doe3[5][4:20],list_doe3[16][4:20],list_doe3[27][4:20]]            # Construction type
        RvalRoof    = str2fl([list_doe3[6][4:20],list_doe3[17][4:20],list_doe3[28][4:20]])     # [m2*K/W] R-value
        Uwindow     = str2fl([list_doe3[7][4:20],list_doe3[18][4:20],list_doe3[29][4:20]])     # [W/m2*K] U-factor
        SHGC        = str2fl([list_doe3[8][4:20],list_doe3[19][4:20],list_doe3[30][4:20]])     # [-] coefficient
        HVAC        = str2fl([list_doe3[9][4:20],list_doe3[20][4:20],list_doe3[31][4:20]])     # [kW] Air Conditioning
        HEAT        = str2fl([list_doe3[10][4:20],list_doe3[21][4:20],list_doe3[32][4:20]])    # [kW] Heating
        COP         = str2fl([list_doe3[11][4:20],list_doe3[22][4:20],list_doe3[33][4:20]])    # [-] Air Conditioning COP
        EffHeat     = str2fl([list_doe3[12][4:20],list_doe3[23][4:20],list_doe3[34][4:20]])    # [%] Heating Efficiency
        FanFlow     = str2fl([list_doe3[13][4:20],list_doe3[24][4:20],list_doe3[35][4:20]])    # [m3/s] Fan Max Flow Rate

        # Read Schedules (Sheet 4)
        file_doe_name_schedules = os.path.join("{}".format(DIR_DOE_PATH), "BLD{}".format(i+1),"BLD{}_Schedules.csv".format(i+1))
        list_doe4 = read_csv(file_doe_name_schedules)

        #listof(listof weekday, sat, sun (list of 24 fractions)))
        SchEquip    = str2fl([list_doe4[1][6:30],list_doe4[2][6:30],list_doe4[3][6:30]])      # Equipment Schedule 24 hrs
        SchLight    = str2fl([list_doe4[4][6:30],list_doe4[5][6:30],list_doe4[6][6:30]])      # Light Schedule 24 hrs; Wkday=Sat=Sun=Hol
        SchOcc      = str2fl([list_doe4[7][6:30],list_doe4[8][6:30],list_doe4[9][6:30]])      # Occupancy Schedule 24 hrs
        SetCool     = str2fl([list_doe4[10][6:30],list_doe4[11][6:30],list_doe4[12][6:30]])   # Cooling Setpoint Schedule 24 hrs
        SetHeat     = str2fl([list_doe4[13][6:30],list_doe4[14][6:30],list_doe4[15][6:30]])   # Heating Setpoint Schedule 24 hrs; summer design
        SchGas      = str2fl([list_doe4[16][6:30],list_doe4[17][6:30],list_doe4[18][6:30]])   # Gas Equipment Schedule 24 hrs; wkday=sat
        SchSWH      = str2fl([list_doe4[19][6:30],list_doe4[20][6:30],list_doe4[21][6:30]])   # Solar Water Heating Schedule 24 hrs; wkday=summerdesign, sat=winterdesgin


        for j in range(3):

            # j = 3 built eras
            #print"\tEra: {} @j={}".format(BUILTERA[j], j)

            for k in range(16):

                # k = 16 climate zones
                #print "\tClimate zone: {} @k={}".format(ZONETYPE[k], k)

                B = Building(
                    hCeiling[j],                        # floorHeight by era
                    1,                                  # intHeatNight
                    1,                                  # intHeatDay
                    0.1,                                # intHeatFRad
                    0.1,                                # intHeatFLat
                    Infil[j],                           # infil (ACH) by era
                    Vent[j]/1000.,                      # vent (m^3/s/m^2) by era, converted from liters
                    glazing[j],                         # glazing ratio by era
                    Uwindow[j][k],                      # uValue by era, by climate type
                    SHGC[j][k],                         # SHGC, by era, by climate type
                    'AIR',                              # cooling condensation system type: AIR, WATER
                    COP[j][k],                          # cop by era, climate type
                    297,                                # coolSetpointDay = 24 C
                    297,                                # coolSetpointNight
                    293,                                # heatSetpointDay = 20 C
                    293,                                # heatSetpointNight
                    (HVAC[j][k]*1000.0)/AreaFloor[j],   # coolCap converted to W/m2 by era, climate type
                    EffHeat[j][k],                      # heatEff by era, climate type
                    293)                                # initialTemp at 20 C

                #Not defined in the constructor
                B.heatCap = (HEAT[j][k]*1000.0)/AreaFloor[j]         # heating Capacity converted to W/m2 by era, climate type
                B.Type = BLDTYPE[i]
                B.Era = BUILTERA[j]
                B.Zone = ZONETYPE[k]
                refDOE[i][j][k] = B

                # Define wall, mass(floor), roof
                # Reference from E+ for conductivity, thickness (reference below)

                # Material: (thermalCond, volHeat = specific heat * density)
                Concrete = Material (1.311, 836.8 * 2240,"Concrete")
                Insulation = Material (0.049, 836.8 * 265.0, "Insulation")
                Gypsum = Material (0.16, 830.0 * 784.9, "Gypsum")
                Wood = Material (0.11, 1210.0 * 544.62, "Wood")
                Stucco = Material(0.6918,  837.0 * 1858.0, "Stucco")

                # Wall (1 in stucco, concrete, insulation, gypsum)
                # Check TypWall by era, by climate
                if TypeWall[j][k] == "MassWall":
                    #Construct wall based on R value of Wall from refDOE and properties defined above
                    # 1" stucco, 8" concrete, tbd insulation, 1/2" gypsum
                    Rbase = 0.271087 # R val based on stucco, concrete, gypsum
                    Rins = RvalWall[j][k] - Rbase #find insulation value
                    D_ins = Rins * Insulation.thermalCond # depth of ins from m2*K/W * W/m*K = m
                    if D_ins > 0.01:
                        thickness = [0.0254,0.0508,0.0508,0.0508,0.0508,D_ins,0.0127]
                        layers = [Stucco,Concrete,Concrete,Concrete,Concrete,Insulation,Gypsum]
                    else:
                        #if it's less then 1 cm don't include in layers
                        thickness = [0.0254,0.0508,0.0508,0.0508,0.0508,0.0127]
                        layers = [Stucco,Concrete,Concrete,Concrete,Concrete,Gypsum]

                    wall = Element(0.08,0.92,thickness,layers,0.,293.,0.,"MassWall")

                    # If mass wall, assume mass floor (4" concrete)
                    # Mass (assume 4" concrete);
                    alb = 0.2
                    emis = 0.9
                    thickness = [0.054,0.054]
                    concrete = Material (1.31, 2240.0*836.8)
                    mass = Element(alb,emis,thickness,[concrete,concrete],0,293,1,"MassFloor")

                elif TypeWall[j][k] == "WoodFrame":
                    # 0.01m wood siding, tbd insulation, 1/2" gypsum
                    Rbase = 0.170284091    # based on wood siding, gypsum
                    Rins = RvalWall[j][k] - Rbase
                    D_ins = Rins * Insulation.thermalCond #depth of insulatino

                    if D_ins > 0.01:
                        thickness = [0.01,D_ins,0.0127]
                        layers = [Wood,Insulation,Gypsum]
                    else:
                        thickness = [0.01,0.0127]
                        layers = [Wood,Gypsum]

                    wall = Element(0.22,0.92,thickness,layers,0.,293.,0.,"WoodFrameWall")

                    # If wood frame wall, assume wooden floor
                    alb = 0.2
                    emis = 0.9
                    thickness = [0.05,0.05]
                    wood = Material(1.31, 2240.0*836.8)
                    mass = Element(alb,emis,thickness,[wood,wood],0.,293.,1.,"WoodFloor")

                elif TypeWall[j][k] == "SteelFrame":
                    # 1" stucco, 8" concrete, tbd insulation, 1/2" gypsum
                    Rbase = 0.271087 # based on stucco, concrete, gypsum
                    Rins = RvalWall[j][k] - Rbase
                    D_ins = Rins * Insulation.thermalCond
                    if D_ins > 0.01:
                        thickness = [0.0254,0.0508,0.0508,0.0508,0.0508,D_ins,0.0127]
                        layers = [Stucco,Concrete,Concrete,Concrete,Concrete,Insulation,Gypsum]
                    else:    # If insulation is too thin, assume no insulation
                        thickness = [0.0254,0.0508,0.0508,0.0508,0.0508,0.0127]
                        layers = [Stucco,Concrete,Concrete,Concrete,Concrete,Gypsum]
                    wall = Element(0.15,0.92,thickness,layers,0.,293.,0.,"SteelFrame")

                    # If mass wall, assume mass foor
                    # Mass (assume 4" concrete),
                    alb = 0.2
                    emis = 0.93
                    thickness = [0.05,0.05]
                    mass = Element(alb,emis,thickness,[Concrete,Concrete],0.,293.,1.,"MassFloor")

                elif TypeWall[j][k] == "MetalWall":
                    # metal siding, insulation, 1/2" gypsum
                    alb = 0.2
                    emis = 0.9
                    D_ins = max((RvalWall[j][k] * Insulation.thermalCond)/2, 0.01) #use derived insul thickness or 0.01 based on max
                    thickness = [D_ins,D_ins,0.0127]
                    materials = [Insulation,Insulation,Gypsum]
                    wall = Element(alb,emis,thickness,materials,0,293,0,"MetalWall")

                    # Mass (assume 4" concrete);
                    alb = 0.2
                    emis = 0.9
                    thickness = [0.05, 0.05]
                    concrete = Material(1.31, 2240.0*836.8)
                    mass = Element(alb,emis,thickness,[concrete,concrete],0.,293.,1.,"MassFloor")

                # Roof
                if TypeRoof[j][k] == "IEAD": #Insulation Entirely Above Deck
                    # IEAD-> membrane, insulation, decking
                     alb = 0.2
                     emis = 0.93
                     D_ins = max(RvalRoof[j][k] * Insulation.thermalCond/2.,0.01);
                     roof = Element(alb,emis,[D_ins,D_ins],[Insulation,Insulation],0.,293.,0.,"IEAD")

                elif TypeRoof[j][k] == "Attic":
                    # IEAD-> membrane, insulation, decking
                    alb = 0.2
                    emis = 0.9
                    D_ins = max(RvalRoof[j][k] * Insulation.thermalCond/2.,0.01)
                    roof = Element(alb,emis,[D_ins,D_ins],[Insulation,Insulation],0.,293.,0.,"Attic")

                elif TypeRoof[j][k] == "MetalRoof":
                    # IEAD-> membrane, insulation, decking
                    alb = 0.2
                    emis = 0.9
                    D_ins = max(RvalRoof[j][k] * Insulation.thermalCond/2.,0.01)
                    roof = Element(alb,emis,[D_ins,D_ins],[Insulation,Insulation],0.,293.,0.,"MetalRoof")

                # Define bulding energy model, set fraction of the urban floor space of this typology to zero
                refBEM[i][j][k] = BEMDef(B, mass, wall, roof, 0.0)
                refBEM[i][j][k].building.FanMax = FanFlow[j][k] # max fan flow rate (m^3/s) per DOE

                Schedule[i][j][k] = SchDef()

                Schedule[i][j][k].Elec = SchEquip   # 3x24 matrix of schedule for fraction electricity (WD,Sat,Sun)
                Schedule[i][j][k].Light = SchLight  # 3x24 matrix of schedule for fraction light (WD,Sat,Sun)
                Schedule[i][j][k].Gas = SchGas      # 3x24 matrix of schedule for fraction gas (WD,Sat,Sun)
                Schedule[i][j][k].Occ = SchOcc      # 3x24 matrix of schedule for fraction occupancy (WD,Sat,Sun)
                Schedule[i][j][k].Cool = SetCool    # 3x24 matrix of schedule for fraction cooling temp (WD,Sat,Sun)
                Schedule[i][j][k].Heat = SetHeat    # 3x24 matrix of schedule for fraction heating temp (WD,Sat,Sun)
                Schedule[i][j][k].SWH = SchSWH      # 3x24 matrix of schedule for fraction SWH (WD,Sat,Sun

                Schedule[i][j][k].Qelec = Elec[j]                   # W/m^2 (max) for electrical plug process
                Schedule[i][j][k].Qlight = Light[j]                 # W/m^2 (max) for light
                Schedule[i][j][k].Nocc = Occupant[j]/AreaFloor[j]   # Person/m^2
                Schedule[i][j][k].Qgas = Gas[j]                     # W/m^2 (max) for gas
                Schedule[i][j][k].Vent = Vent[j]/1000.0             # m^3/m^2 per person
                Schedule[i][j][k].Vswh = SHW[j]/AreaFloor[j]        # litres per hour per m^2 of floor


    # if not test serialize refDOE,refBEM,Schedule and store in resources
    if serialize_output:

        # create a binary file for serialized obj
        pkl_file_path = os.path.join(DIR_CURR,'refdata','readDOE.pkl')
        pickle_readDOE = open(pkl_file_path, 'wb')

        # dump in ../resources
        # Pickle objects, protocol 1 b/c binary file
        pickle.dump(refDOE, pickle_readDOE,1)
        pickle.dump(refBEM, pickle_readDOE,1)
        pickle.dump(Schedule, pickle_readDOE,1)

        pickle_readDOE.close()

    return refDOE, refBEM, Schedule

if __name__ == "__main__":

    # Set to True only if you want create new .pkls of DOE refs
    # Use --serialize switch to serialize the readDOE data
    if len(sys.argv)> 1 and sys.argv[1]=="--serialize":
        refDOE, refBEM, Schedule = readDOE(True)
    else:
        refDOE, refBEM, Schedule = readDOE(False)


# Material ref from E+
#     1/2IN Gypsum,            !- Name
#     Smooth,                  !- Roughness
#     0.0127,                  !- Thickness {m}
#     0.1600,                  !- Conductivity {W/m-K}
#     784.9000,                !- Density {kg/m3}
#     830.0000,                !- Specific Heat {J/kg-K}
#     0.9000,                  !- Thermal Absorptance
#     0.9200,                  !- Solar Absorptance
#     0.9200;                  !- Visible Absorptance
#
# Material,
#     1IN Stucco,              !- Name
#     Smooth,                  !- Roughness
#     0.0253,                  !- Thickness
#     0.6918,                  !- Conductivity
#     1858.0000,               !- Density
#     837.0000,                !- Specific Heat
#     0.9000,                  !- Thermal Absorptance
#     0.9200,                  !- Solar Absorptance
#     0.9200;                  !- Visible Absorptance
#
# Material,
#     8IN CONCRETE HW,  !- Name
#     Rough,                   !- Roughness
#     0.2032,                  !- Thickness {m}
#     1.3110,                  !- Conductivity {W/m-K}
#     2240.0000,               !- Density {kg/m3}
#     836.8000,                !- Specific Heat {J/kg-K}
#     0.9000,                  !- Thermal Absorptance
#     0.7000,                  !- Solar Absorptance
#     0.7000;                  !- Visible Absorptance
#
# Material,
#     Mass NonRes Wall Insulation, !- Name
#     MediumRough,             !- Roughness
#     0.0484268844343858,      !- Thickness {m}
#     0.049,                   !- Conductivity {W/m-K}
#     265.0000,                !- Density {kg/m3}
#     836.8000,                !- Specific Heat {J/kg-K}
#     0.9000,                  !- Thermal Absorptance
#     0.7000,                  !- Solar Absorptance
#     0.7000;                  !- Visible Absorptance
#
# Material,
#     Std Wood 6inch,          !- Name
#     MediumSmooth,            !- Roughness
#     0.15,                    !- Thickness {m}
#     0.12,                    !- Conductivity {W/m-K}
#     540.0000,                !- Density {kg/m3}
#     1210,                    !- Specific Heat {J/kg-K}
#     0.9000000,               !- Thermal Absorptance
#     0.7000000,               !- Solar Absorptance
#     0.7000000;               !- Visible Absorptance! Common Materials
#
# Material,
#     Wood Siding,             !- Name
#     MediumSmooth,            !- Roughness
#     0.0100,                  !- Thickness {m}
#     0.1100,                  !- Conductivity {W/m-K}
#     544.6200,                !- Density {kg/m3}
#     1210.0000,               !- Specific Heat {J/kg-K}
#     0.9000,                  !- Thermal Absorptance
#     0.7800,                  !- Solar Absorptance
#     0.7800;                  !- Visible Absorptance



from __future__ import division, print_function

try:
    range = xrange
except NameError:
    pass

from math import sqrt, pow
import copy


class UCMDef(object):
    """
    Definition of Urban Canopy - Building Energy Model Class

    properties
        road;          % Road element class (moved from BEM)

        % Urban Canyon Parameters
        bldHeight;     % average building height (m)
        bldDensity;    % horizontal building density (footprint)
        verToHor;      % vertical-to-horizontal urban area ratio (facade area/urban area)
        treeCoverage;  % horizontal tree density (footprint)
        sensAnthrop;   % sensible anthropogenic heat (other than from buildings) (W m-2)
        latAnthrop;    % latent anthropogenic heat (other than from buildings) (W m-2)
        z0u;           % urban roughness length (m)
        l_disp;          % urban displacement length (m)
        roadShad;      % shadowing of roads
        canWidth;      % canyon width (m)
        bldWidth;      % bld width (m)
        canAspect;     % canyon aspect ratio
        roadConf;      % road-sky configuration factors (sky view factor)
        alb_wall;      % average wall albedo
        wallConf;      % wall-sky configuration factors (sky view factor)
        VFwallroad;    % wall-road view factor
        VFroadwall;    % road-wall view factor
        facArea;       % facade area (m2)
        roadArea;      % road area (m2)
        roofArea;      % roof area (m2) (also building area)
        facAbsor;      % average facade absortivity
        roadAbsor;     % average road absortivity
        h_mix;         % waste heat mix into canyon ratio

        % Urban Canyon Variables
        canTemp;       % canyon air temperature (db) (K)
        Tdp;           % dew point temperature
        Twb;           % wetbulb temperature
        canHum;        % canyon specific humidity (kg kg-1)
        canRHum;       % canyon relative humidity (%)
        canWind;       % urban canyon wind velocity (m s-1)
        turbU;         % canyon turbulent velocities (m s-1)
        turbV;         % canyon turbulent velocities (m s-1)
        turbW;         % canyon turbulent velocities (m s-1)
        ublTemp;       % urban boundary layer temperature (K)
        ublTempdx;     % urban boundary layer temperature discretization (K)
        ublWind;       % urban boundary layer wind velocity (m s-1)
        ustar;         % friction velocity (m s-1)
        ustarMod;      % modified friction velocity (m s-1)
        uExch;         % exchange velocity (m s-1)
        treeLatHeat;   % latent heat from trees (W m-2)
        treeSensHeat;  % sensible heat from trees (W m-2)
        sensHeat;      % urban sensible heat (W m-2)
        latHeat;       % urban latent heat (W m-2)
        windProf;      % urban wind profile
        Q_roof;        % sensible heat flux from building roof (convective)
        Q_wall;        % sensible heat flux from building wall (convective)
        Q_window;      % sensible heat flux from building window (via U-factor)
        Q_road;        % sensible heat flux from road (convective)
        Q_hvac;        % sensible heat flux from HVAC waste
        Q_traffic;     % sensible heat flux from traffic (net)
        Q_ubl;         % Convective heat exchange with UBL layer
        Q_vent;        % Convective heat exchange from ventilation/infiltration
        SolRecWall;    % Solar received by wall
        SolRecRoof;    % Solar received by roof
        SolRecRoad;    % Solar received by road
        roadTemp;      % average road temperature (K)
        roofTemp;      % average roof temperature (K)
        wallTemp;      % average wall temperature (K)
        ElecTotal;     % Total Electricity consumption of urban area
        GasTotal;      % Total Gas consumption of the urban area
    """
    CANYON_TEMP_BOUND_ERROR = "Something obviously went wrong (UCMDef.py)... "

    def __init__(self,bldHeight,bldDensity,verToHor,treeCoverage,sensAnthrop,latAnthrop,
        initialTemp,initialHum,initialWind,parameter,r_glaze,SHGC,alb_wall,road):
        self.road = road                                            # Road element class (moved from BEM)
        self.bldHeight = bldHeight                                  # average building height (m)
        self.verToHor = verToHor                                    # vertical-to-horizontal urban area ratio (facade area/urban area)
        self.bldDensity = bldDensity                                # horizontal building density (footprint/total_area)
        self.treeCoverage = treeCoverage                            # horizontal tree density (footprint)
        self.sensAnthrop = sensAnthrop                              # sensible anthropogenic heat (other than from buildings) (W m-2)
        self.latAnthrop = latAnthrop                                # latent anthropogenic heat (other than from buildings) (W m-2)
        self.roadShad = min(treeCoverage/(1-bldDensity),1)          # fraction of road not building shadowed
        # Key to understanding next few formulas is that UWG_Matlab assumes bld_area = square, so sqrt(bld_area) = side length
        self.bldWidth = 4*bldHeight*bldDensity/verToHor             # bld width (side length) derived from bldDensity and verToHor (m)
        d = self.bldWidth/(sqrt(bldDensity))                        # urban area width == sqrt(bldDensity) == ratio of bld footprint_width / urban area footprint
        self.canWidth = d - self.bldWidth                           # canyon width (m) = urban area width - building width
        self.canAspect = bldHeight/self.canWidth                    # canyon aspect ratio
        self.roadConf = pow(pow(self.canAspect,2)+1,0.5) - self.canAspect   # road-sky configuration factor (sky view factor SVF)
        self.wallConf = 0.5 * (self.canAspect + 1 -
            pow(pow(self.canAspect,2)+1,0.5)) / (self.canAspect)    # wall-sky configuration factor (sky view factor SVF)
        self.facArea = 4*self.bldWidth*bldHeight                    # bld width [m]
        self.roadArea = d*d - pow(self.bldWidth,2)                  # road area [m2] = urban_area_ - bld_area
        self.roofArea = pow(self.bldWidth,2)                        # roof area [m2]
        self.canTemp = initialTemp                                  # canyon air temperature (db) (K)
        self.roadTemp = initialTemp                                 # average road temperature (K)
        self.canHum = initialHum                                    # canyon specific humidity (kg kg-1)
        self.ublWind = max(initialWind,parameter.windMin)           # urban boundary layer wind volocity (m s-1)
        self.canWind = initialWind                                  # urban canyon wind velocity (m s-1)
        self.ustar = 0.1*initialWind                                # friction velocity (m s-1)
        self.ustarMod = 0.1*initialWind                             # modified friction velocity (m s-1)

        # Calculate z0u = urban roughness length (m)
        frontDens = verToHor/4.                                     # density of just street facing facade
        if frontDens < 0.15:
          self.z0u = frontDens * self.bldHeight
        else:
          self.z0u = 0.15 * self.bldHeight

        # Calculate l_dsp = urban displacement length (m)
        if frontDens < 0.05:
          self.l_disp = 3 * frontDens * self.bldHeight
        elif frontDens < 0.15:
          self.l_disp = (0.15+5.5*(frontDens-0.05))*self.bldHeight
        elif frontDens < 1:
          self.l_disp = (0.7+0.35*(frontDens-0.15))*self.bldHeight
        else:
          self.l_disp = 0.5*self.bldHeight

        self.alb_wall = alb_wall                                    # average wall albedo (-)
        self.facAbsor = (1-r_glaze)*(1-alb_wall) + r_glaze*(1-0.75*SHGC) # avg facade absorptivity (-) == wall_mat_fraction * absorption + window_frac * non_solar_heat_gain
        self.roadAbsor = (1-road.vegCoverage)*(1-road.albedo)       # average road absorptivity
        self.sensHeat = 0.0                                         # urban sensible heat [W m-2]
        # Variables set in urbflux()
        self.latHeat = None                                         # urban latent heat [W m-2]
        self.windProf = []                                          # wind profile
        self.canRHum = None
        self.Tdp = None

    def __repr__(self):
        return "UCMDef: ver2Hor = {}, bldDens = {}, canyon H/W = {}/{}, canyon aspect ratio = {}, bld width = {}, roof Area = {}".format(
            self.verToHor,
            self.bldDensity,
            int(self.bldHeight),
            int(self.canWidth),
            round(self.canAspect,1),
            self.facArea,
            self.roofArea
            )

    def UCModel(self,BEM,T_ubl,forc,parameter):
        # Calculate the urban canyon temperature per The uwg (2012) Eq. 10
        dens = forc.pres/(1000*0.287042*self.canTemp*(1.+1.607858*self.canHum))     # air density
        dens_ubl = forc.pres/(1000*0.287042*T_ubl*(1.+1.607858*forc.hum))           # air density
        Cp_air = parameter.cp

        self.Q_wall = 0.
        self.Q_window = 0.
        self.Q_road = 0.
        self.Q_hvac = 0.
        self.Q_traffic = 0.
        self.Q_vent = 0.
        self.Q_ubl = 0.
        self.ElecTotal = 0.
        self.GasTotal = 0.
        self.roofTemp = 0.
        self.wallTemp = 0.

        # Road to Canyon
        T_road = self.road.layerTemp[0]
        h_conv = self.road.aeroCond
        H1 = T_road*h_conv*self.roadArea       # Heat (Sens) from road surface
        H2 = h_conv*self.roadArea

        H1 = H1 + T_ubl*self.roadArea*self.uExch*Cp_air*dens_ubl # Heat from UBL
        H2 = H2 + self.roadArea*self.uExch*Cp_air*dens_ubl
        Q = (self.roofArea+self.roadArea)*(self.sensAnthrop + self.treeSensHeat*self.treeCoverage)

        # Building energy output to canyon, in terms of absolute (total) values
        for j in range(len(BEM)):
            # Re-naming variable for readability
            building = BEM[j].building
            wall = BEM[j].wall
            T_indoor = building.indoorTemp
            T_wall = wall.layerTemp[0]
            R_glazing = building.glazingRatio
            A_wall = (1.-R_glazing)*self.facArea
            A_window = R_glazing*self.facArea
            U_window = building.uValue

            H1 = H1 + BEM[j].frac*(                 # fraction of the urban floor space of this typology
                T_indoor*A_window*U_window +        # window U
                T_wall*A_wall*h_conv +              # Wall conv
                T_indoor*self.roofArea*BEM[j].building.vent*BEM[j].building.nFloor*Cp_air*dens +            # Vent
                T_indoor*self.roofArea*BEM[j].building.infil*self.bldHeight/3600.0*Cp_air*dens)             # Infil

            H2 = H2 + BEM[j].frac*(
                A_window*U_window +
                A_wall*h_conv +
                self.roofArea*BEM[j].building.vent*BEM[j].building.nFloor*Cp_air*dens +                     # Vent
                self.roofArea*BEM[j].building.infil*self.bldHeight/3600.0*Cp_air*dens)                      # Infil

            Q = Q + BEM[j].frac*(
                self.roofArea*building.sensWaste*self.h_mix +           # HVAC waste heat
                A_window*BEM[j].wall.solRec*(1.0-BEM[j].building.shgc))   # heat that didn't make it to inside

            self.wallTemp = self.wallTemp + BEM[j].frac*T_wall
            self.roofTemp = self.roofTemp + BEM[j].frac*BEM[j].roof.layerTemp[0]
            self.Q_ubl = self.Q_ubl + BEM[j].frac*self.bldDensity*(BEM[j].roof.sens + BEM[j].building.sensWaste*(1.-self.h_mix)) # Changed by Jiachen Mao in March 2017

        # Solve for canyon temperature
        self.canTemp = (H1 + Q)/H2

        # Heat flux based per m^2 of urban area
        self.Q_road = h_conv*(T_road-self.canTemp)*(1.-self.bldDensity)  # Sensible heat from road (W/m^2 of urban area)
        self.Q_ubl = self.Q_ubl + self.uExch*Cp_air*dens*(self.canTemp-T_ubl)*(1.-self.bldDensity)
        self.Q_wall = h_conv*(self.wallTemp-self.canTemp)*(self.verToHor)
        self.Q_traffic = self.sensAnthrop

        # Building energy output to canyon, per m^2 of urban area
        T_can = copy.copy(self.canTemp)
        for j in range(len(BEM)):
            V_vent = BEM[j].building.vent*BEM[j].building.nFloor  # ventilation volume per m^2 of building
            V_infil = BEM[j].building.infil*self.bldHeight/3600.0
            T_indoor = BEM[j].building.indoorTemp
            U_window = BEM[j].building.uValue                     # Added by Jiachen Mao in March 2017
            R_glazing = BEM[j].building.glazingRatio              # Changed by Jiachen Mao in March 2017

            self.Q_window = self.Q_window + BEM[j].frac*self.verToHor*R_glazing*U_window*(T_indoor-T_can)
            self.Q_window = self.Q_window + BEM[j].frac*self.verToHor*R_glazing*BEM[j].wall.solRec*(1.-BEM[j].building.shgc)
            self.Q_vent = self.Q_vent + BEM[j].frac*self.bldDensity*Cp_air*dens*(V_vent + V_infil)*(T_indoor-T_can)
            self.Q_hvac = self.Q_hvac + BEM[j].frac*self.bldDensity*BEM[j].building.sensWaste*self.h_mix

            self.Q_roof = self.Q_roof + BEM[j].frac*self.bldDensity*BEM[j].roof.sens

            # Total Electrical & Gas power in MW
            self.ElecTotal = self.ElecTotal + BEM[j].fl_area*BEM[j].building.ElecTotal/1.e6
            self.GasTotal = self.GasTotal + BEM[j].fl_area*BEM[j].building.GasTotal/1.e6

        # Sensible Heat
        # N.B In the current uwg code, latent heat from evapotranspiration, stagnant water,
        # or anthropogenic sources is not modelled due to the difficulty of validation, and
        # lack of reliability of precipitation data from EPW files.
        self.sensHeat = self.Q_wall + self.Q_road + self.Q_vent + self.Q_window + self.Q_hvac + self.Q_traffic + self.treeSensHeat + self.Q_roof

        # Error checking
        if self.canTemp > 350. or self.canTemp < 250:
            raise Exception(self.CANYON_TEMP_BOUND_ERROR)



def infracalcs(UCM,forc,e_road,e_wall,T_road,T_wall):
    # Calculate IR surface flux
    sigma = 5.67e-8                         # Stephen-Boltzman const
    road_wall_conf = (1. - UCM.roadConf)    # configuration factor (view factor) for road to wall
    wall_road_conf = UCM.wallConf           # wall to road VF same as wall-sky configuration factors (sky view factor)

    # Calculate radiation of unshaded road, accounting for radiation exchange from wall
    infra_road = e_road * UCM.roadConf * (1.-UCM.roadShad) * (forc.infra - sigma*T_road**4.) + (1.-UCM.roadShad) * e_wall * e_road * sigma * road_wall_conf * (T_wall**4.-T_road**4.)
    # Calculate radiation of wall, accounting for radiation exchange from unshaded road
    infra_wall = e_wall * UCM.wallConf * (forc.infra - sigma*T_wall**4.) + (1.-UCM.roadShad) * e_wall * e_road * sigma * wall_road_conf * (T_road**4.-T_wall**4.)

    return infra_road, infra_wall



from __future__ import division, print_function

try:
    range = xrange
except NameError:
    pass

import math


class SimParam(object):
    """
    SimParam
    Calculates simulation time parameters based on initial date
    and weather data time step, and simulation timestep

    properties
        dt            % Simulation time-step
        timeForcing   % Weather data time-step
        month         % Begin month
        day           % Begin day of the month
        days          % Number of days of simulation
        timePrint     % time-step for printing outputs
        timeDay       % number of timesteps in a design-day
        timeSim       % number of timesteps of the simulation
        timeMax       % total simulation time (s)
        nt            % total number of timesteps
        timeFinal     % final timestep of simulation
        timeInitial   % initial timestep of simulation
        secDay        % seconds of one day (s)
        hourDay       % hour of the day (0 - 23hr)
        inobis        % julian day at the end of each month
        julian        % julian day
    """
    TIMESTEP_CONFLICT_MSG = "TIMESTEP ERROR! Timestep must be a factor of 3600."

    def __init__(self,dt,timefor,M,DAY,days):
        self.dt = dt                                                        # uwg time simulation time step
        self.timeForcing = timefor                                          # weather data timestep
        self.month = int(M)
        self.day = DAY
        self.days = days
        self.timePrint = timefor                                            # weather data timestep
        self.timeDay = 24*3600/timefor                                      # how many times weather senses in a day
        self.timeSim = self.timeDay*days                                    # how many steps in weather data simulation
        self.timeMax = 24.*3600.*days                                       # total seconds in simulation days
        self.nt = int(round(self.timeMax/self.dt+1))                        # total number of timesteps for uwg simuation
        self.inobis = [0,31,59,90,120,151,181,212,243,273,304,334]
        self.julian = self.inobis[self.month - 1] + DAY - 1
        #H1: (julian day * number of timesteps in a day) == sensor data index in epw
        H1 = int((self.inobis[self.month - 1] + DAY - 1) * self.timeDay)
        self.timeInitial = H1 + 8                                           # sensor data in epw for intial time based on julian day & timesteps
        self.timeFinal = int(H1 + self.timeDay * self.days - 1 + 8)         # sensor data in epw for final time based on julian day & timesteps
        self.secDay = 0                                                     # current seconds in day
        self.hourDay = 0                                                    # current hours in day

    def __repr__(self):
        return "SimParam: Start = {}/{}, days = {}, timestep = {}s".format(
            self.month,
            int(self.day),
            self.days,
            self.dt
            )

    def is_near_zero(self,num,eps=1e-10):
        return abs(float(num)) < eps

    def UpdateDate(self):
        self.secDay = self.secDay + self.dt

        if self.is_near_zero(self.secDay - 3600*24):
            self.day += 1
            self.julian = self.julian + 1
            self.secDay = 0.
            for j in range(12):
                if self.is_near_zero(self.julian - self.inobis[j]):
                    self.month = self.month + 1
                    self.day = 1

        if self.secDay > (3600*24):
            raise Exception("{}. CURRENTLY AT {}.".format(self.TIMESTEP_CONFLICT_MSG, self.dt))

        self.hourDay = int(math.floor(self.secDay/3600.))       # 0 - 23hr



from __future__ import division

try:
    range = xrange
except NameError:
    pass

import math


class Element(object):
    """
    uwg Element

    # Note: In matlab not all instance variables are instantiated. They are assumed to be a 0-by-0 empty matrix
    # https://www.mathworks.com/help/matlab/matlab_oop/specifying-properties.html

    Attributes:
        albedo;          % outer surface albedo
        emissivity;      % outer surface emissivity
        layerThickness;  % vector of layer thicknesses (m)
        layerThermalCond;% vector of layer thermal conductivities (W m-1 K-1)
        layerVolHeat;    % vector of layer volumetric heat (J m-3 K-1)
        vegCoverage;     % surface vegetation coverage
        layerTemp;       % vector of layer temperatures (K)
        waterStorage;    % thickness of water film (m) (only for horizontal surfaces)
        horizontal;      % 1-horizontal, 0-vertical
        solRec;          % solar radiation received (W m-2)
        infra;           % net longwave radiation (W m-2)
        lat;             % surface latent heat flux (W m-2)
        sens;            % surface sensible heat flux (W m-2)
        solAbs;          % solar radiation absorbed (W m-2)
        aeroCond;        % convective heat transfer
        T_ext;           % external surface temperature
        T_int;           % internal surface temperature
        flux;            % external surface heat flux
    """

    THICKNESSLST_EQ_MATERIALLST_MSG = \
    "-----------------------------------------\n" +\
    "ERROR: the number of layer thickness must\n" +\
    "match the number of layer materials\n"
    "-----------------------------------------"
    CONDUCTION_INPUT_MSG = 'ERROR: check input parameters in the Conduction routine'

    def __init__(self, alb, emis, thicknessLst, materialLst, vegCoverage, T_init, horizontal,name=None):
        if len(thicknessLst) != len(materialLst):
            raise Exception(self.THICKNESSLST_EQ_MATERIALLST_MSG)
        else:
            self._name = name                                       # purely for internal process
            self.albedo = alb                                       # outer surface albedo
            self.emissivity = emis                                  # outer surface emissivity
            self.layerThickness = thicknessLst                      # vector of layer thicnesses (m)
            self.layerThermalCond = [0. for i in materialLst]       # vector of layer thermal conductivity (W m-1 K-1)
            self.layerVolHeat = [0. for i in materialLst]           # vector of layer volumetric heat (J m-3 K-1)

            # Create list of layer k and (Cp*density) from materialLst properties
            for i in range(len(materialLst)):
              self.layerThermalCond[i] = materialLst[i].thermalCond
              self.layerVolHeat[i] = materialLst[i].volHeat

            self.vegCoverage = vegCoverage                          # surface vegetation coverage
            self.layerTemp = [T_init] * len(thicknessLst)           # vector of layer temperatures (K)
            self.waterStorage = 0.                                  # thickness of water film (m) for horizontal surfaces only
            self.infra = 0.                                         # net longwave radiation (W m-2)
            self.horizontal = horizontal                            # 1-horizontal, 0-vertical
            self.sens = 0.                                          # surface sensible heat flux (W m-2)

            # B/c we have to explicity define this in python. Set as None
            self.solRec = None                                       # solar radiation received (W m-2)
            self.lat = None                                          # surface latent heat flux (W m-2)
            self.solAbs = None                                       # solar radiation absorbed (W m-2)
            self.aeroCond = None                                     # convective heat transfer
            self.T_ext = None                                        # external surface temperature
            self.T_int = None                                        # internal surface temperature
            self.flux = None                                         # external surface heat flux

    def __repr__(self):
        # Returns some representative wall properties
        s1 = "Element: {a}\n\tlayerNum={b}, totaldepth={c}\n\t".format(
            a=self._name,
            b=len(self.layerThickness),
            c=sum(self.layerThickness)
            )
        s2 = "e={d}, a={e}\n\tr_val={f}, Cp*dens_avg={g}\n\tlayerTemp: {h}".format(
            d=self.emissivity,
            e=self.albedo,
            f=round(sum(self.layerThermalCond)/2.,2),
            g=round(sum(self.layerVolHeat)/2.,2),
            h=self.layerTemp
            )
        return s1 + s2

    def is_near_zero(self,num,eps=1e-10):
        return abs(float(num)) < eps

    def SurfFlux(self,forc,parameter,simTime,humRef,tempRef,windRef,boundCond,intFlux):
        """ Calculate net heat flux, and update element layer temperatures
        """

        # Calculated per unit area (m^2)
        dens = forc.pres/(1000*0.287042*tempRef*(1.+1.607858*humRef)) # air density (kgd m-3)
        self.aeroCond = 5.8 + 3.7 * windRef         # Convection coef (ref: uwg, eq. 12))

        if (self.horizontal):     # For roof, mass, road
            # Evaporation (m s-1), Film water & soil latent heat
            if not self.is_near_zero(self.waterStorage) and self.waterStorage > 0.0:
                # N.B In the current uwg code, latent heat from evapotranspiration, stagnant water,
                # or anthropogenic sources is not modelled due to the difficulty of validation, and
                # lack of reliability of precipitation data from EPW files.Therefore this condition
                # is never run because all elements have had their waterStorage hardcoded to 0.
                qtsat = self.qsat([self.layerTemp[0]],[forc.pres],parameter)[0]
                eg = self.aeroCond*parameter.colburn*dens*(qtsat-humRef)/parameter.waterDens/parameter.cp
                self.waterStorage = min(self.waterStorage + simTime.dt*(forc.prec-eg),parameter.wgmax)
                self.waterStorage = max(self.waterStorage,0.) # (m)
            else:
                eg = 0.
            soilLat = eg*parameter.waterDens*parameter.lv

            # Winter, no veg
            if simTime.month < parameter.vegStart and simTime.month > parameter.vegEnd:
                self.solAbs = (1.-self.albedo)*self.solRec # (W m-2)
                vegLat = 0.
                vegSens = 0.
            else:    # Summer, veg
                self.solAbs = ((1.-self.vegCoverage)*(1.-self.albedo)+self.vegCoverage*(1.-parameter.vegAlbedo))*self.solRec
                vegLat = self.vegCoverage*parameter.grassFLat*(1.-parameter.vegAlbedo)*self.solRec
                vegSens = self.vegCoverage*(1.-parameter.grassFLat)*(1.-parameter.vegAlbedo)*self.solRec

            self.lat = soilLat + vegLat
            # Sensible & net heat flux
            self.sens = vegSens + self.aeroCond*(self.layerTemp[0]-tempRef)
            self.flux = -self.sens + self.solAbs + self.infra - self.lat # (W m-2)

        else:               # For vertical surfaces (wall)
            self.solAbs = (1.-self.albedo)*self.solRec
            self.lat = 0.

            # Sensible & net heat flux
            self.sens = self.aeroCond*(self.layerTemp[0]-tempRef)
            self.flux = -self.sens + self.solAbs + self.infra - self.lat # (W m-2)

        self.layerTemp = self.Conduction(simTime.dt, self.flux, boundCond, forc.deepTemp, intFlux)
        self.T_ext = self.layerTemp[0]
        self.T_int = self.layerTemp[-1]

    def Conduction(self, dt, flx1, bc, temp2, flx2):
        """
        Solve the conductance of heat based on of the element layers.
        arg:
            flx1  : net heat flux on surface
            bc    : boundary condition parameter (1 or 2)
            temp2 : deep soil temperature (ave of air temperature)
            flx2  : surface flux (sum of absorbed, emitted, etc.)

        key prop:
            za = [[ x00, x01, x02 ... x0w ]
                  [ x10, x11, x12 ... x1w ]
                            ...
                  [ xh0, xh1, xh2 ... xhw ]]

            where h = matrix row index    = element layer number
                  w = matrix column index = 3

        """
        t = self.layerTemp          # vector of layer temperatures (K)
        hc = self.layerVolHeat      # vector of layer volumetric heat (J m-3 K-1)
        tc = self.layerThermalCond  # vector of layer thermal conductivities (W m-1 K-1)
        d = self.layerThickness     # vector of layer thicknesses (m)

        # flx1                      : net heat flux on surface
        # bc                        : boundary condition parameter (1 or 2)
        # temp2                     : deep soil temperature (avg of air temperature)
        # flx2                      : surface flux (sum of absorbed, emitted, etc.)

        fimp = 0.5                  # implicit coefficient
        fexp = 0.5                  # explicit coefficient
        num = len(t)                # number of layers

        # Mean thermal conductivity over distance between 2 layers (W/mK)
        tcp = [0 for x in range(num)]
        # Thermal capacity times layer depth (J/m2K)
        hcp = [0 for x in range(num)]
        # lower, main, and upper diagonals
        za = [[0 for y in range(3)] for x in range(num)]
        # RHS
        zy = [0 for x in range(num)]

        #--------------------------------------------------------------------------
        # Define the column vectors for heat capactiy and conductivity
        hcp[0] = hc[0] * d[0]
        for j in range(1,num):
            tcp[j] = 2. / (d[j-1] / tc[j-1] + d[j] / tc[j])
            hcp[j] = hc[j] * d[j]

        #--------------------------------------------------------------------------
        # Define the first row of za matrix, and RHS column vector
        za[0][0] = 0.
        za[0][1] = hcp[0]/dt + fimp*tcp[1]
        za[0][2] = -fimp*tcp[1]
        zy[0] = hcp[0]/dt*t[0] - fexp*tcp[1]*(t[0]-t[1]) + flx1

        #--------------------------------------------------------------------------
        # Define other rows
        for j in range(1,num-1):
          za[j][0] = fimp*(-tcp[j])
          za[j][1] = hcp[j]/dt + fimp*(tcp[j]+tcp[j+1])
          za[j][2] = fimp*(-tcp[j+1])
          zy[j] = hcp[j]/dt * t[j] + fexp * \
            (tcp[j]*t[j-1] - tcp[j]*t[j] - tcp[j+1]*t[j] + tcp[j+1]*t[j+1])

        #--------------------------------------------------------------------------
        # Boundary conditions
        if self.is_near_zero(bc-1.): # heat flux
            za[num-1][0] = fimp * (-tcp[num-1])
            za[num-1][1] = hcp[num-1]/dt + fimp*tcp[num-1]
            za[num-1][2] = 0.
            zy[num-1] = hcp[num-1]/dt*t[num-1] + fexp*tcp[num-1]*(t[num-2]-t[num-1]) + flx2
        elif self.is_near_zero(bc-2.): # deep-temperature
            za[num-1][0] = 0.
            za[num-1][1] = 1.
            za[num-1][2] = 0.
            zy[num-1] = temp2
        else:
            raise Exception(self.CONDUCTION_INPUT_MSG)

        #--------------------------------------------------------------------------

        zx = self.invert(num,za,zy)
        #t(:) = zx(:);
        return zx # return zx as 1d vector of templayers

    def qsat(self,temp,pres,parameter):
        """
        Calculate (qsat_lst) vector of saturation humidity from:
            temp = vector of element layer temperatures
            pres = pressure (at current timestep).
        """
        gamw = (parameter.cl - parameter.cpv) / parameter.rv
        betaw = (parameter.lvtt/parameter.rv) + (gamw * parameter.tt)
        alpw = math.log(parameter.estt) + (betaw /parameter.tt) + (gamw * math.log(parameter.tt))
        work2 = parameter.r/parameter.rv
        foes_lst = [0 for i in range(len(temp))]
        work1_lst = [0 for i in range(len(temp))]
        qsat_lst = [0 for i in range(len(temp))]

        for i in range(len(temp)):
          # saturation vapor pressure
          foes_lst[i] = math.exp( alpw - betaw/temp[i] - gamw*math.log(temp[i])  )
          work1_lst[i] = foes_lst[i]/pres[i]
          # saturation humidity
          qsat_lst[i] = work2*work1_lst[i] / (1. + (work2-1.) * work1_lst[i])

        return qsat_lst


    def invert(self,nz,A,C):
        """
        Inversion and resolution of a tridiagonal matrix
                 A X = C
        Input:
         nz number of layers
         a(*,1) lower diagonal (Ai,i-1)
         a(*,2) principal diagonal (Ai,i)
         a(*,3) upper diagonal (Ai,i+1)
         c
        Output
         x     results
        """

        X = [0 for i in range(nz)]

        for i in reversed(range(nz-1)):
            C[i] = C[i] - A[i][2] * C[i+1]/A[i+1][1]
            A[i][1] = A[i][1] - A[i][2] * A[i+1][0]/A[i+1][1]

        for i in  range(1,nz,1):
            C[i] = C[i] - A[i][0] * C[i-1]/A[i-1][1]

        for i in range(nz):
            X[i] = C[i]/A[i][1]

        return X



"""Collection of useful methods."""
import os
from csv import reader as csv_reader
import sys

try:
    range = xrange
except NameError:
    pass


def zeros(h, w):
    """create a (h x w) matrix of zeros.

    Args:
        h: Height of the matrix.
        w: Width of the matrix.
    """
    return [[0 for x in range(w)] for y in range(h)]


def read_csv(file_name_):
    # open csv file and read
    if os.path.exists(file_name_):
        if sys.version_info[0] >= 3:
            file_ = open(file_name_, "r", errors='ignore')
        else:
            file_ = open(file_name_, "r")

        gen_ = csv_reader(file_, delimiter=",")
        L = [r for r in gen_]
        file_.close()
        return L
    else:
        raise Exception("File name: '{}' does not exist.".format(file_name_))


def is_near_zero(num, eps=1e-10):
    return abs(num) < eps


def str2fl(x):
    """Recurses through lists and converts lists of string to float

    Args:
        x: string or list of strings
    """
    def helper_to_fl(s_):
        """ deals with odd string imports converts to float"""
        if s_ == "":
            return "null"
        elif "," in s_:
            s_ = s_.replace(",", "")

        try:
            return float(s_)
        except:
            return (s_)

    fl_lst = []
    if isinstance(x[0], str):                # Check if list of strings, then sent to conversion
        for xi in range(len(x)):
            fl_lst.append(helper_to_fl(x[xi]))
    elif isinstance(x[0], list):                    # Check if list of lists, then recurse
        for xi in range(len(x)):
            fl_lst.append(str2fl(x[xi]))
    else:
        return False

    return fl_lst



from __future__ import division

try:
    range = xrange
except NameError:
    pass

import logging


class UBLDef(object):
    """
    %   Class definition for the Urban Boundary Layer (UBL)

    properties
        location;       % relative location within a city (N,NE,E,SE,S,SW,W,NW,C)
        charLength;     % characteristic length of the urban area (m)
        perimeter;      % urban area perimeter (m)
        urbArea;        % horizontal urban area (m2)
        orthLength;     % length of the side of the urban area orthogonal
                        % to the wind direction (m)
        paralLength;    % length of the side of the urban area parallel
                        % to the wind direction (m)
        ublTemp;        % urban boundary layer temperature (K)
        ublTempdx;      % urban boundary layer temperature discretization (K)
        advHeat;        % advection heat flux (W m-2)
        sensHeat;       % sensible heat flux (W m-2)
        dayBLHeight;    % daytime mixing height, orig = 700
        nightBLHeight;  % Sing: 80, Bub-Cap: 50, nighttime boundary-layer height (m); orig 80
    end
    """

    def __init__(self,location,charLength,initialTemp,maxdx,dayBLHeight,nightBLHeight):
        self.location = location                                    # relative location within a city (N,NE,E,SE,S,SW,W,NW,C)
        self.charLength = charLength                                # characteristic length of the urban area (m)
        self.perimeter = 4. * charLength
        self.urbArea = charLength**2                                # horizontal urban area (m2)
        self.orthLength = charLength                                # length of the side of the urban area orthogonal
                                                                    # to the wind direction (m)
        numdx = round(charLength/min(charLength,maxdx))
        self.paralLength = charLength/numdx                         # length of the side of the urban area parallel
                                                                    # to the wind direction (m)
        self.ublTemp = initialTemp                                  # urban boundary layer temperature (K)
        self.ublTempdx = [initialTemp for x in range(int(numdx))]   # urban boundary layer temperature discretization (K)
        self.dayBLHeight = dayBLHeight                              # daytime mixing height, orig = 700
        self.nightBLHeight = nightBLHeight                          # Sing: 80, Bub-Cap: 50, nighttime boundary-layer height (m); orig 80

        # Logger will be disabled by default unless explicitly called in tests
        self.logger = logging.getLogger(__name__)

    def __repr__(self):
        return "UBL: urbArea {}m2, charLength {}m".format(
            self.urbArea,
            self.charLength
            )

    def is_near_zero(self, num, eps=1e-14):
        return abs(float(num)) < eps

    def UBLModel(self,UCM,RSM,rural,forc,parameter,simTime):
        # Note that only one urban canyon area is considered
        self.sensHeat = UCM.sensHeat
        heatDif = max(self.sensHeat - rural.sens,0)
        Cp = parameter.cp                           # Heat capacity of air (J/kg.K)
        k_w = parameter.circCoeff                   # k_w per Bueno 'the uwg', eq 8
        g = parameter.g                             # Gravity
        v_wind = max(forc.wind,parameter.windMin)   # wind velocity

        # Air density
        refDens = 0.
        for iz in range(RSM.nzref):
            refDens = refDens + RSM.densityProfC[iz] * RSM.dz[iz] / (RSM.z[RSM.nzref-1] + RSM.dz[RSM.nzref-1]/2.)

        forDens = 0.
        for iz in range(RSM.nzfor):
            forDens = forDens + RSM.densityProfC[iz] * RSM.dz[iz] / (RSM.z[RSM.nzfor-1] + RSM.dz[RSM.nzfor-1]/2.)

        # ---------------------------------------------------------------------
        # Day
        # ---------------------------------------------------------------------
        time = simTime.secDay/3600.
        noon = 12.
        daylimit = parameter.dayThreshold      # sunlight threshold for day (~150W/m^2)
        nightlimit = parameter.dayThreshold    # sunlight threshold for night (~50W/m^2)
        sunlight = forc.dir + forc.dif

        # If dir & dif light is greater than threshold, use day
        is_day = (sunlight > daylimit) and (time < noon or self.is_near_zero(time-noon)) \
                or (sunlight > nightlimit) and (time > noon) or (self.sensHeat > 150.0)
        if is_day:
            # Circulation velocity per Bueno 'the uwg', eq 8
            self.logger.debug("{} Day ubl calcs".format(__name__))
            h_UBL = self.dayBLHeight            # Day boundary layer height
            eqTemp = RSM.tempProf[RSM.nzref-1]
            eqWind = RSM.windProf[RSM.nzref-1]

            Csurf = UCM.Q_ubl*simTime.dt/(h_UBL*refDens*Cp)
            u_circ = k_w*(g*heatDif/Cp/refDens/eqTemp*h_UBL)**(1./3.)

            if v_wind > u_circ:   # Forced problem (usually this)
                advCoef  = self.orthLength*eqWind*simTime.dt/self.urbArea*1.4
                self.ublTemp = (Csurf + advCoef * eqTemp + self.ublTemp)/(1. + advCoef)
                self.ublTempdx = [self.ublTemp for x in range(len(self.ublTempdx))]
            else:                   # Convective problem
                advCoef  = self.perimeter*u_circ*simTime.dt/self.urbArea*1.4
                self.ublTemp = (Csurf+advCoef*eqTemp + self.ublTemp)/(1 + advCoef)
                self.ublTempdx = [self.ublTemp for x in range(len(self.ublTempdx))]



        # ---------------------------------------------------------------------
        # Night
        # ---------------------------------------------------------------------
        else:
            self.logger.debug("{} Night ubl calcs".format(__name__))
            h_UBL = self.nightBLHeight      # Night boundary layer height
            Csurf = UCM.Q_ubl*simTime.dt/(h_UBL*refDens*Cp)
            self.ublTemp, self.ublTempdx = self.NightForc(self.ublTempdx,simTime.dt, \
                h_UBL,self.paralLength,self.charLength,RSM,Csurf)

        self.logger.debug("ublTemp = {}".format(self.ublTemp))

    def NightForc(self,ublTempdx,dt,h_UBL,paralLength,charLength,RSM,Csurf):
        # Night forcing (RSM.nzfor = number of layers of forcing)
        # Average potential temperature & wind speed of the profile
        intAdv1 = 0.
        for iz in range(RSM.nzfor):
            intAdv1 = intAdv1 + RSM.windProf[iz] * RSM.tempProf[iz] * RSM.dz[iz]

        advCoef1 = 1.4*dt/paralLength/h_UBL*intAdv1

        intAdv2 = 0
        for iz in range(RSM.nzfor):
            intAdv2 = intAdv2 + RSM.windProf[iz]*RSM.dz[iz]

        advCoef2 = 1.4*dt/paralLength/h_UBL*intAdv2

        ublTempdx[0] = (Csurf + advCoef1 + ublTempdx[0])/(1 + advCoef2)
        ublTemp = ublTempdx[0]

        for i in range(1,int(charLength)//int(paralLength)):
            eqTemp = ublTempdx[i-1]
            ublTempdx[i] = (Csurf + advCoef2*eqTemp + ublTempdx[i])/(1 + advCoef2)
            ublTemp = ublTemp + ublTempdx[i]

        # ublTemp/charLength*paralLength;
        ublTemp = ublTemp/float(charLength)*float(paralLength)

        return ublTemp, ublTempdx



"""Urban Weather Generator Library."""

__version__ = '5.3.2'

from .simparam import SimParam
from .weather import Weather
from .building import Building
from .material import Material
from .element import Element
from .BEMDef import BEMDef
from .schdef import SchDef
from .param import Param
from .UCMDef import UCMDef
from .forcing import Forcing
from .UBLDef import UBLDef
from .RSMDef import RSMDef
from .solarcalcs import SolarCalcs

from .readDOE import readDOE
from .infracalcs import infracalcs
from .urbflux import urbflux

from .uwg import uwg
from .uwg import procMat


__all__ = [
    "uwg",
    "utilities",
    "material",
    "element",
    "building",
    "BEMDef",
    "forcing",
    "param",
    "psychrometrics",
    "schdef",
    "simparam",
    "UCMDef",
    "urbflux",
    "weather",
    "RSMDef",
    ]



"""
=========================================================================
 THE URBAN WEATHER GENERATOR (uwg)
=========================================================================
Version 4.2

Original Author: B. Bueno
Edited by A. Nakano & Lingfu Zhang
Modified by Joseph Yang (joeyang@mit.edu) - May, 2016
Translated to Python by Saeran Vasanthakumar - February, 2018

Original Pulbication on the uwg's Methods:
Bueno, Bruno; Norford, Leslie; Hidalgo, Julia; Pigeon, Gregoire (2013).
The urban weather generator, Journal of Building Performance Simulation. 6:4,269-281.
doi: 10.1080/19401493.2012.718797
=========================================================================
"""
from __future__ import division, print_function
from functools import reduce

try:
    range = xrange
except NameError:
    pass

import os
import math
import copy
import logging

try:
    import cPickle as pickle
except ImportError:
    import pickle

from .simparam import SimParam
from .weather import Weather
from .building import Building
from .material import Material
from .element import Element
from .BEMDef import BEMDef
from .schdef import SchDef
from .param import Param
from .UCMDef import UCMDef
from .forcing import Forcing
from .UBLDef import UBLDef
from .RSMDef import RSMDef
from .solarcalcs import SolarCalcs
from .psychrometrics import psychrometrics
from .readDOE import readDOE
from .urbflux import urbflux
from . import utilities

# For debugging only
#from pprint import pprint
#from decimal import Decimal
#pp = pprint
#dd = Decimal.from_float


class uwg(object):
    """Morph a rural EPW file to urban conditions using a file with a list of urban parameters.

    args:
        epwDir: The directory in which the rural EPW file sits.
        epwFileName: The name of the rural epw file that will be morphed.
        uwgParamDir: The directory in which the uwg Parameter File (.uwg) sits.
        uwgParamFileName: The name of the uwg Parameter File (.uwg).
        destinationDir: Optional destination directory for the morphed EPW file.
            If left blank, the morphed file will be written into the same directory
            as the rural EPW file (the epwDir).
        destinationFileName: Optional destination file name for the morphed EPW file.
            If left blank, the morphed file will append "_UWG" to the original file name.
    returns:
        newClimateFile: the path to a new EPW file that has been morphed to account
            for uban conditions.
    """

    """ Section 1 - Definitions for constants / other parameters """
    MINTHICKNESS = 0.01    # Minimum layer thickness (to prevent crashing) (m)
    MAXTHICKNESS = 0.05    # Maximum layer thickness (m)
    # http://web.mit.edu/parmstr/Public/NRCan/nrcc29118.pdf (Figly & Snodgrass)
    SOILTCOND = 1
    # http://www.europment.org/library/2013/venice/bypaper/MFHEEF/MFHEEF-21.pdf (average taken from Table 1)
    SOILVOLHEAT = 2e6
    # Soil material used for soil-depth padding
    SOIL = Material(SOILTCOND, SOILVOLHEAT, name="soil")

    # Physical constants
    G = 9.81               # gravity (m s-2)
    CP = 1004.             # heat capacity for air (J/kg K)
    VK = 0.40              # von karman constant (dimensionless)
    R = 287.               # gas constant dry air (J/kg K)
    RV = 461.5             # gas constant water vapor (J/kg K)
    LV = 2.26e6            # latent heat of evaporation (J/kg)
    SIGMA = 5.67e-08       # Stefan Boltzmann constant (W m-2 K-4)
    WATERDENS = 1000.      # water density (kg m-3)
    LVTT = 2.5008e6        #
    TT = 273.16            #
    ESTT = 611.14          #
    CL = 4.218e3           #
    CPV = 1846.1           #
    B = 9.4                # Coefficients derived by Louis (1979)
    CM = 7.4               #
    # (Pr/Sc)^(2/3) for Colburn analogy in water evaporation
    COLBURN = math.pow((0.713/0.621), (2/3.))

    # Site-specific parameters
    WGMAX = 0.005  # maximum film water depth on horizontal surfaces (m)

    # File path parameter
    RESOURCE_PATH = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "resources"))
    CURRENT_PATH = os.path.abspath(os.path.dirname(__file__))

    def __init__(self, epwFileName, uwgParamFileName=None, epwDir=None, uwgParamDir=None, destinationDir=None, destinationFileName=None):

        # Logger will be disabled by default unless explicitly called in tests
        self.logger = logging.getLogger(__name__)

        # User defined
        self.epwFileName = epwFileName if epwFileName.lower().endswith('.epw') else epwFileName + \
            '.epw'  # Revise epw file name if not end with epw
        # If file name is entered then will uwg will set input from .uwg file
        self.uwgParamFileName = uwgParamFileName

        # If user does not overload
        self.destinationFileName = destinationFileName if destinationFileName else self.epwFileName.strip(
            '.epw') + '_UWG.epw'
        self.epwDir = epwDir if epwDir else os.path.join(self.RESOURCE_PATH, "epw")
        self.uwgParamDir = uwgParamDir if uwgParamDir else os.path.join(
            self.RESOURCE_PATH, "parameters")
        self.destinationDir = destinationDir if destinationDir else os.path.join(
            self.RESOURCE_PATH, "epw_uwg")

        # refdata: Serialized DOE reference data, z_meso height data
        self.readDOE_file_path = os.path.join(self.CURRENT_PATH, "refdata", "readDOE.pkl")
        self.z_meso_dir_path = os.path.join(self.CURRENT_PATH, "refdata")

        # EPW precision
        self.epw_precision = 1

        # init uwg variables
        self._init_param_dict = None

        # Define Simulation and Weather parameters
        self.Month = None        # starting month (1-12)
        self.Day = None          # starting day (1-31)
        self.nDay = None         # number of days
        self.dtSim = None        # simulation time step (s)
        self.dtWeather = None    # seconds (s)

        # HVAC system and internal laod
        self.autosize = None     # autosize HVAC (1 or 0)
        self.sensOcc = None      # Sensible heat from occupant
        self.LatFOcc = None      # Latent heat fraction from occupant (normally 0.3)
        self.RadFOcc = None      # Radiant heat fraction from occupant (normally 0.2)
        self.RadFEquip = None    # Radiant heat fraction from equipment (normally 0.5)
        self.RadFLight = None    # Radiant heat fraction from light (normally 0.7)

        # Define Urban microclimate parameters
        self.h_ubl1 = None       # ubl height - day (m)
        self.h_ubl2 = None       # ubl height - night (m)
        self.h_ref = None        # inversion height
        self.h_temp = None       # temperature height
        self.h_wind = None       # wind height
        self.c_circ = None       # circulation coefficient
        self.c_exch = None       # exchange coefficient
        self.maxDay = None       # max day threshhold
        self.maxNight = None     # max night threshhold
        self.windMin = None      # min wind speed (m/s)
        self.h_obs = None        # rural average obstacle height

        # Urban characteristics
        self.bldHeight = None    # average building height (m)
        self.h_mix = None        # mixing height (m)
        self.bldDensity = None   # building density (0-1)
        self.verToHor = None     # building aspect ratio
        # radius defining the urban area of study [aka. characteristic length] (m)
        self.charLength = None
        self.alb_road = None     # road albedo
        self.d_road = None       # road pavement thickness
        self.sensAnth = None     # non-building sensible heat (W/m^2)
        self.latAnth = None      # non-building latent heat heat (W/m^2). Not used, taken out by JH.


        # Fraction of building typology stock
        self.bld = None         # 16x3 matrix of fraction of building type by era

        # climate Zone
        self.zone = None

        # Vegetation parameters
        self.vegCover = None     # urban area veg coverage ratio
        self.treeCoverage = None  # urban area tree coverage ratio
        self.vegStart = None     # vegetation start month
        self.vegEnd = None       # vegetation end month
        self.albVeg = None       # Vegetation albedo
        self.rurVegCover = None  # rural vegetation cover
        self.latGrss = None      # latent fraction of grass
        self.latTree = None      # latent fraction of tree

        # Define Traffic schedule
        self.SchTraffic = None

        # Define Road (Assume 0.5m of asphalt)
        self.kRoad = None       # road pavement conductivity (W/m K)
        self.cRoad = None       # road volumetric heat capacity (J/m^3 K)

        # Define optional Building characteristics
        self.flr_h = None       # floor-to-floor height
        self.albRoof = None     # roof albedo (0 - 1)
        self.vegRoof = None     # Fraction of the roofs covered in grass/shrubs (0-1)
        self.glzR = None        # Glazing Ratio
        self.SHGC = None       # Solar Heat Gain Coefficient
        self.albWall = None    # Wall albedo

    def ToString(self):
        """Overwrite .NET ToString method."""
        return self.__repr__()

    def __repr__(self):
        def _split_string(s):
            return s[0] + ":\n  " + s[1].replace(",", "\n  ")

        def _tabbed(s):
            return _split_string(s.__repr__().split(":"))

        def _list_2_tabbed(b):
            return reduce(lambda a, b: a+"\n"+b, [_tabbed(_b) for _b in b])

        return "uwg for {}:\n\n{}{}{}{}{}{}{}{}".format(
            self.epwFileName,
            _tabbed(self.simTime)+"\n" if hasattr(self, "simTime") else "No simTime attr.\n",
            _tabbed(self.weather)+"\n" if hasattr(self, "weather") else "No weather attr.\n",
            _tabbed(self.geoParam)+"\n" if hasattr(self, "geoParam") else "No geoParam attr.\n",
            _tabbed(self.UBL)+"\n" if hasattr(self, "UBL") else "No UBL attr.\n",
            "Rural "+_tabbed(self.RSM)+"\n" if hasattr(self, "RSM") else "No Rural RSM attr.\n",
            "Urban "+_tabbed(self.USM)+"\n" if hasattr(self, "USM") else "No Urban RSM attr.\n",
            _tabbed(self.UCM)+"\n" if hasattr(self, "UCM") else "No UCM attr.\n",
            _list_2_tabbed(self.BEM) if hasattr(self, "BEM") else "No BEM attr."
            )

    def is_near_zero(self,num,eps=1e-10):
        return abs(float(num)) < eps

    def read_epw(self):
        """Section 2 - Read EPW file
        properties:
            self.climateDataPath
            self.newPathName
            self._header    # header data
            self.epwinput   # timestep data for weather
            self.lat        # latitude
            self.lon        # longitude
            self.GMT        # GMT
            self.nSoil      # Number of soil depths
            self.Tsoil      # nSoil x 12 matrix for soil temperture (K)
            self.depth_soil # nSoil x 1 matrix for soil depth (m)
        """

        # Make dir path to epw file
        self.climateDataPath = os.path.join(self.epwDir, self.epwFileName)

        # Open epw file and feed csv data to climate_data
        try:
            climate_data = utilities.read_csv(self.climateDataPath)
        except Exception as e:
            raise Exception("Failed to read epw file! {}".format(e.message))

        # Read header lines (1 to 8) from EPW and ensure TMY2 format.
        self._header = climate_data[0:8]

        # Read weather data from EPW for each time step in weather file. (lines 8 - end)
        self.epwinput = climate_data[8:]

        # Read Lat, Long (line 1 of EPW)
        self.lat = float(self._header[0][6])
        self.lon = float(self._header[0][7])
        self.GMT = float(self._header[0][8])

        # Read in soil temperature data (assumes this is always there)
        # ref: http://bigladdersoftware.com/epx/docs/8-2/auxiliary-programs/epw-csv-format-inout.html
        soilData = self._header[3]
        self.nSoil = int(soilData[1])           # Number of ground temperature depths
        self.Tsoil = utilities.zeros(self.nSoil, 12)  # nSoil x 12 matrix for soil temperture (K)
        self.depth_soil = utilities.zeros(self.nSoil, 1)   # nSoil x 1 matrix for soil depth (m)

        # Read monthly data for each layer of soil from EPW file
        for i in range(self.nSoil):
            self.depth_soil[i][0] = float(soilData[2 + (i*16)])  # get soil depth for each nSoil
            # Monthly data
            for j in range(12):
                # 12 months of soil T for specific depth
                self.Tsoil[i][j] = float(soilData[6 + (i*16) + j]) + 273.15

        # Set new directory path for the moprhed EPW file
        self.newPathName = os.path.join(self.destinationDir, self.destinationFileName)

    def read_input(self):
        """Section 3 - Read Input File (.m, file)
        Note: UWG_Matlab input files are xlsm, XML, .m, file.
        properties:
            self._init_param_dict   # dictionary of simulation initialization parameters

            self.sensAnth           # non-building sensible heat (W/m^2)
            self.SchTraffic         # Traffice schedule

            self.BEM                # list of BEMDef objects extracted from readDOE
            self.Sch                # list of Schedule objects extracted from readDOE

        """

        uwg_param_file_path = os.path.join(self.uwgParamDir, self.uwgParamFileName)

        if not os.path.exists(uwg_param_file_path):
            raise Exception("Param file: '{}' does not exist.".format(uwg_param_file_path))

        # Open .uwg file and feed csv data to initializeDataFile
        try:
            uwg_param_data = utilities.read_csv(uwg_param_file_path)
        except Exception as e:
            raise Exception("Failed to read .uwg file! {}".format(e.message))

        # The initialize.uwg is read with a dictionary so that users changing
        # line endings or line numbers doesn't make reading input incorrect
        self._init_param_dict = {}
        count = 0
        while count < len(uwg_param_data):
            row = uwg_param_data[count]
            row = [row[i].replace(" ", "") for i in range(len(row))]  # strip white spaces

            # Optional parameters might be empty so handle separately
            is_optional_parameter = (
                row != [] and
                (
                    row[0] == "albRoof" or
                    row[0] == "vegRoof" or
                    row[0] == "glzR" or
                    row[0] == "hvac" or
                    row[0] == "albWall" or
                    row[0] == "SHGC"
                )
            )
            try:
                if row == [] or "#" in row[0]:
                    count += 1
                    continue
                elif row[0] == "SchTraffic":
                    # SchTraffic: 3 x 24 matrix
                    trafficrows = uwg_param_data[count+1:count+4]
                    self._init_param_dict[row[0]] = [utilities.str2fl(r[:24]) for r in trafficrows]
                    count += 4
                elif row[0] == "bld":
                    # bld: 17 x 3 matrix
                    bldrows = uwg_param_data[count+1:count+17]
                    self._init_param_dict[row[0]] = [utilities.str2fl(r[:3]) for r in bldrows]
                    count += 17
                elif is_optional_parameter:
                    self._init_param_dict[row[0]] = float(row[1]) if row[1] != "" else None
                    count += 1
                else:
                    self._init_param_dict[row[0]] = float(row[1])
                    count += 1
            except ValueError:
                print("Error while reading parameter at {} {}".format(count, row))

        ipd = self._init_param_dict

        # Define Simulation and Weather parameters
        if self.Month is None: self.Month = ipd['Month']
        if self.Day is None: self.Day = ipd['Day']
        if self.nDay is None: self.nDay = ipd['nDay']
        if self.dtSim is None: self.dtSim = ipd['dtSim']
        if self.dtWeather is None: self.dtWeather = ipd['dtWeather']

        # HVAC system and internal laod
        if self.autosize is None: self.autosize = ipd['autosize']
        if self.sensOcc is None: self.sensOcc = ipd['sensOcc']
        if self.LatFOcc is None: self.LatFOcc = ipd['LatFOcc']
        if self.RadFOcc is None: self.RadFOcc = ipd['RadFOcc']
        if self.RadFEquip is None: self.RadFEquip = ipd['RadFEquip']
        if self.RadFLight is None: self.RadFLight = ipd['RadFLight']

        # Define Urban microclimate parameters
        if self.h_ubl1 is None: self.h_ubl1 = ipd['h_ubl1']
        if self.h_ubl2 is None: self.h_ubl2 = ipd['h_ubl2']
        if self.h_ref is None: self.h_ref = ipd['h_ref']
        if self.h_temp is None: self.h_temp = ipd['h_temp']
        if self.h_wind is None: self.h_wind = ipd['h_wind']
        if self.c_circ is None: self.c_circ = ipd['c_circ']
        if self.c_exch is None: self.c_exch = ipd['c_exch']
        if self.maxDay is None: self.maxDay = ipd['maxDay']
        if self.maxNight is None: self.maxNight = ipd['maxNight']
        if self.windMin is None: self.windMin = ipd['windMin']
        if self.h_obs is None: self.h_obs = ipd['h_obs']

        # Urban characteristics
        if self.bldHeight is None: self.bldHeight = ipd['bldHeight']
        if self.h_mix is None: self.h_mix = ipd['h_mix']
        if self.bldDensity is None: self.bldDensity = ipd['bldDensity']
        if self.verToHor is None: self.verToHor = ipd['verToHor']
        if self.charLength is None: self.charLength = ipd['charLength']
        if self.alb_road is None: self.alb_road = ipd['albRoad']
        if self.d_road is None: self.d_road = ipd['dRoad']
        if self.sensAnth is None: self.sensAnth = ipd['sensAnth']
        # if self.latAnth is None: self.latAnth = ipd['latAnth'] # Not used, taken out by JH.

        # climate Zone
        if self.zone is None: self.zone = ipd['zone']

        # Vegetation parameters
        if self.vegCover is None: self.vegCover = ipd['vegCover']
        if self.treeCoverage is None: self.treeCoverage = ipd['treeCoverage']
        if self.vegStart is None: self.vegStart = ipd['vegStart']
        if self.vegEnd is None: self.vegEnd = ipd['vegEnd']
        if self.albVeg is None: self.albVeg = ipd['albVeg']
        if self.rurVegCover is None: self.rurVegCover = ipd['rurVegCover']
        if self.latGrss is None: self.latGrss = ipd['latGrss']
        if self.latTree is None: self.latTree = ipd['latTree']

        # Define Traffic schedule
        if self.SchTraffic is None: self.SchTraffic = ipd['SchTraffic']

        # Define Road (Assume 0.5m of asphalt)
        if self.kRoad is None: self.kRoad = ipd['kRoad']
        if self.cRoad is None: self.cRoad = ipd['cRoad']

        # Building stock fraction
        if self.bld is None: self.bld = ipd['bld']

        # Optional parameters
        if self.albRoof is None: self.albRoof = ipd['albRoof']
        if self.vegRoof is None: self.vegRoof = ipd['vegRoof']
        if self.glzR is None: self.glzR = ipd['glzR']
        if self.albWall is None: self.albWall = ipd['albWall']
        if self.SHGC is None: self.SHGC = ipd['SHGC']

    def check_required_inputs(self):
        # Fail if required parameters aren't correct
        assert isinstance(self.Month, (float, int)), \
            'Month must be a number. Got {}'.format(type(self.Month))
        assert isinstance(self.Day, (float, int)), \
            'Day must be a number. Got {}'.format(type(self.Day))
        assert isinstance(self.nDay, (float, int)), \
            'nDay must be a number. Got {}'.format(type(self.nDay))
        assert isinstance(self.dtSim, float), \
            'dtSim must be a float. Got {}'.format(type(self.dtSim))
        assert isinstance(self.dtWeather, float), \
            'dtWeather must be a float. Got {}'.format(type(self.dtWeather))
        assert isinstance(self.autosize, (float, int)), \
            'autosize must be a number. Got {}'.format(type(self.autosize))
        assert isinstance(self.sensOcc, float), \
            'sensOcc must be a float. Got {}'.format(type(self.sensOcc))
        assert isinstance(self.LatFOcc, float), \
            'LatFOcc must be a float. Got {}'.format(type(self.LatFOcc))
        assert isinstance(self.RadFOcc, float), \
            'RadFOcc must be a float. Got {}'.format(type(self.RadFOcc))
        assert isinstance(self.RadFEquip, float), \
            'RadFEquip must be a float. Got {}'.format(type(self.RadFEquip))
        assert isinstance(self.RadFLight, float), \
            'RadFLight must be a float. Got {}'.format(type(self.RadFLight))
        assert isinstance(self.h_ubl1, float), \
            'h_ubl1 must be a float. Got {}'.format(type(self.h_ubl1))
        assert isinstance(self.h_ubl2, float), \
            'h_ubl2 must be a float. Got {}'.format(type(self.h_ubl2))
        assert isinstance(self.h_ref, float), \
            'h_ref must be a float. Got {}'.format(type(self.h_ref))
        assert isinstance(self.h_temp, float), \
            'h_temp must be a float. Got {}'.format(type(self.h_temp))
        assert isinstance(self.h_wind, float), \
            'h_wind must be a float. Got {}'.format(type(self.h_wind))
        assert isinstance(self.c_circ, float), \
            'c_circ must be a float. Got {}'.format(type(self.c_circ))
        assert isinstance(self.c_exch, float), \
            'c_exch must be a float. Got {}'.format(type(self.c_exch))
        assert isinstance(self.maxDay, float), \
            'maxDay must be a float. Got {}'.format(type(self.maxDay))
        assert isinstance(self.maxNight, float), \
            'maxNight must be a float. Got {}'.format(type(self.maxNight))
        assert isinstance(self.windMin, float), \
            'windMin must be a float. Got {}'.format(type(self.windMin))
        assert isinstance(self.h_obs, float), \
            'h_obs must be a float. Got {}'.format(type(self.h_obs))
        assert isinstance(self.bldHeight, float), \
            'bldHeight must be a float. Got {}'.format(type(self.bldHeight))
        assert isinstance(self.h_mix, float), \
            'h_mix must be a float. Got {}'.format(type(self.h_mix))
        assert isinstance(self.bldDensity, float), \
            'bldDensity must be a float. Got {}'.format(type(self.bldDensity))
        assert isinstance(self.verToHor, float), \
            'verToHor must be a float. Got {}'.format(type(self.verToHor))
        assert isinstance(self.charLength, float), \
            'charLength must be a float. Got {}'.format(type(self.charLength))
        assert isinstance(self.alb_road, float), \
            'alb_road must be a float. Got {}'.format(type(self.alb_road))
        assert isinstance(self.d_road, float), \
            'd_road must be a float. Got {}'.format(type(self.d_road))
        assert isinstance(self.sensAnth, float), \
            'sensAnth must be a float. Got {}'.format(type(self.sensAnth))
        # assert isinstance(self.latAnth, float) # Take this out as isn't being used
        assert isinstance(self.bld, list), \
            'bld must be a list. Got {}'.format(type(self.bld))
        assert len(self.bld) == 16, \
            'length of bld must be 16. Got {}'.format(len(self.bld))
        assert isinstance(self.latTree, float), \
            'latTree must be a float. Got {}'.format(type(self.latTree))
        assert isinstance(self.latGrss, float), \
            'latGrss must be a float. Got {}'.format(type(self.latGrss))
        assert isinstance(self.zone, (float, int)), \
            'zone must be a number. Got {}'.format(type(self.zone))
        assert isinstance(self.vegStart, (float, int)), \
            'vegStart must be a number. Got {}'.format(type(self.vegStart))
        assert isinstance(self.vegEnd, (float, int)), \
            'vegEnd must be a number. Got {}'.format(type(self.vegEnd))
        assert isinstance(self.vegCover, float), \
            'vegCover must be a float. Got {}'.format(type(self.vegCover))
        assert isinstance(self.treeCoverage, float), \
            'treeCoverage must be a float. Got {}'.format(type(self.treeCoverage))
        assert isinstance(self.albVeg, float), \
            'albVeg must be a float. Got {}'.format(type(self.albVeg))
        assert isinstance(self.rurVegCover, float), \
            'rurVegCover must be a float. Got {}'.format(type(self.rurVegCover))
        assert isinstance(self.kRoad, float), \
            'kRoad must be a float. Got {}'.format(type(self.kRoad))
        assert isinstance(self.cRoad, float), \
            'cRoad must be a float. Got {}'.format(type(self.cRoad))
        assert isinstance(self.SchTraffic, list), \
            'SchTraffic must be a list. Got {}'.format(type(self.SchTraffic))
        assert len(self.SchTraffic) == 3, \
            'length of SchTraffic must be 3. Got {}'.format(len(self.SchTraffic))

    def set_input(self):
        """ Set inputs from .uwg input file if not already defined, the check if all
        the required input parameters are there.
        """

        # If a uwgParamFileName is set, then read inputs from .uwg file.
        # User-defined class properties will override the inputs from the .uwg file.
        if self.uwgParamFileName is not None:
            print("\nReading uwg file input.")
            self.read_input()
        else:
            print("\nNo .uwg file input.")

        self.check_required_inputs()

        # Modify zone to be used as python index
        self.zone = int(self.zone)-1

    def init_BEM_obj(self):
        """
        Define BEM for each DOE type (read the fraction)
        self.BEM                # list of BEMDef objects
        self.r_glaze            # Glazing ratio for total building stock
        self.SHGC               # SHGC addition for total building stock
        self.alb_wall           # albedo wall addition for total building stock
        """

        if not os.path.exists(self.readDOE_file_path):
            raise Exception("readDOE.pkl file: '{}' does not exist.".format(readDOE_file_path))

        readDOE_file = open(self.readDOE_file_path, 'rb')  # open pickle file in binary form
        refDOE = pickle.load(readDOE_file)
        refBEM = pickle.load(readDOE_file)
        refSchedule = pickle.load(readDOE_file)
        readDOE_file.close()

        # Define building energy models
        k = 0
        self.r_glaze_total = 0.             # Glazing ratio for total building stock
        self.SHGC_total = 0.                # SHGC addition for total building stock
        self.alb_wall_total = 0.            # albedo wall addition for total building stock
        h_floor = self.flr_h or 3.05  # average floor height

        total_urban_bld_area = math.pow(self.charLength, 2)*self.bldDensity * \
            self.bldHeight/h_floor  # total building floor area
        area_matrix = utilities.zeros(16, 3)

        self.BEM = []           # list of BEMDef objects
        self.Sch = []           # list of Schedule objects

        for i in range(16):    # 16 building types
            for j in range(3):  # 3 built eras
                if self.bld[i][j] > 0.:
                    # Add to BEM list
                    self.BEM.append(refBEM[i][j][self.zone])
                    self.BEM[k].frac = self.bld[i][j]
                    self.BEM[k].fl_area = self.bld[i][j] * total_urban_bld_area

                    # Overwrite with optional parameters if provided
                    if self.glzR:
                        self.BEM[k].building.glazingRatio = self.glzR
                    if self.albRoof:
                        self.BEM[k].roof.albedo = self.albRoof
                    if self.vegRoof:
                        self.BEM[k].roof.vegCoverage = self.vegRoof
                    if self.SHGC:
                        self.BEM[k].building.shgc = self.SHGC
                    if self.albWall:
                        self.BEM[k].wall.albedo = self.albWall
                    if self.flr_h:
                        self.BEM[k].building.floorHeight = self.flr_h

                    # Keep track of total urban r_glaze, SHGC, and alb_wall for UCM model
                    self.r_glaze_total += self.BEM[k].frac * self.BEM[k].building.glazingRatio
                    self.SHGC_total += self.BEM[k].frac * self.BEM[k].building.shgc
                    self.alb_wall_total += self.BEM[k].frac * self.BEM[k].wall.albedo
                    # Add to schedule list
                    self.Sch.append(refSchedule[i][j][self.zone])
                    k += 1

    def init_input_obj(self):
        """Section 4 - Create uwg objects from input parameters

            self.simTime            # simulation time parameter obj
            self.weather            # weather obj for simulation time period
            self.forcIP             # Forcing obj
            self.forc               # Empty forcing obj
            self.geoParam           # geographic parameters obj
            self.RSM                # Rural site & vertical diffusion model obj
            self.USM                # Urban site & vertical diffusion model obj
            self.UCM                # Urban canopy model obj
            self.UBL                # Urban boundary layer model

            self.road               # urban road element
            self.rural              # rural road element

            self.soilindex1         # soil index for urban rsoad depth
            self.soilindex2         # soil index for rural road depth
            self.Sch                # list of Schedule objects
        """

        climate_file_path = os.path.join(self.epwDir, self.epwFileName)

        self.simTime = SimParam(self.dtSim, self.dtWeather, self.Month,
                                self.Day, self.nDay)  # simulation time parametrs
        # weather file data for simulation time period
        self.weather = Weather(climate_file_path, self.simTime.timeInitial, self.simTime.timeFinal)
        self.forcIP = Forcing(self.weather.staTemp, self.weather)  # initialized Forcing class
        self.forc = Forcing()  # empty forcing class

        # Initialize geographic Param and Urban Boundary Layer Objects
        nightStart = 18.        # arbitrary values for begin/end hour for night setpoint
        nightEnd = 8.
        maxdx = 250.            # max dx (m)

        self.geoParam = Param(self.h_ubl1, self.h_ubl2, self.h_ref, self.h_temp, self.h_wind, self.c_circ,
                              self.maxDay, self.maxNight, self.latTree, self.latGrss, self.albVeg, self.vegStart, self.vegEnd,
                              nightStart, nightEnd, self.windMin, self.WGMAX, self.c_exch, maxdx, self.G, self.CP, self.VK, self.R,
                              self.RV, self.LV, math.pi, self.SIGMA, self.WATERDENS, self.LVTT, self.TT, self.ESTT, self.CL,
                              self.CPV, self.B, self.CM, self.COLBURN)

        self.UBL = UBLDef(
            'C', self.charLength, self.weather.staTemp[0], maxdx, self.geoParam.dayBLHeight, self.geoParam.nightBLHeight)

        # Defining road
        emis = 0.93
        asphalt = Material(self.kRoad, self.cRoad, 'asphalt')
        road_T_init = 293.
        road_horizontal = 1
        # fraction of surface vegetation coverage
        road_veg_coverage = min(self.vegCover/(1-self.bldDensity), 1.)

        # define road layers
        road_layer_num = int(math.ceil(self.d_road/0.05))
        # 0.5/0.05 ~ 10 x 1 matrix of 0.05 thickness
        thickness_vector = [0.05 for r in range(road_layer_num)]
        material_vector = [asphalt for r in range(road_layer_num)]

        self.road = Element(self.alb_road, emis, thickness_vector, material_vector, road_veg_coverage,
                            road_T_init, road_horizontal, name="urban_road")

        self.rural = copy.deepcopy(self.road)
        self.rural.vegCoverage = self.rurVegCover
        self.rural._name = "rural_road"

        # Reference site class (also include VDM)
        self.RSM = RSMDef(self.lat, self.lon, self.GMT, self.h_obs,
                          self.weather.staTemp[0], self.weather.staPres[0], self.geoParam, self.z_meso_dir_path)
        self.USM = RSMDef(self.lat, self.lon, self.GMT, self.bldHeight/10.,
                          self.weather.staTemp[0], self.weather.staPres[0], self.geoParam, self.z_meso_dir_path)

        T_init = self.weather.staTemp[0]
        H_init = self.weather.staHum[0]

        self.UCM = UCMDef(self.bldHeight, self.bldDensity, self.verToHor, self.treeCoverage, self.sensAnth, self.latAnth, T_init, H_init,
                          self.weather.staUmod[0], self.geoParam, self.r_glaze_total, self.SHGC_total, self.alb_wall_total, self.road)
        self.UCM.h_mix = self.h_mix

        # Define Road Element & buffer to match ground temperature depth
        roadMat, newthickness = procMat(self.road, self.MAXTHICKNESS, self.MINTHICKNESS)

        for i in range(self.nSoil):
            # if soil depth is greater then the thickness of the road
            # we add new slices of soil at max thickness until road is greater or equal

            is_soildepth_equal = self.is_near_zero(self.depth_soil[i][0] - sum(newthickness), 1e-15)

            if is_soildepth_equal or (self.depth_soil[i][0] > sum(newthickness)):
                while self.depth_soil[i][0] > sum(newthickness):
                    newthickness.append(self.MAXTHICKNESS)
                    roadMat.append(self.SOIL)
                self.soilindex1 = i
                break

        self.road = Element(self.road.albedo, self.road.emissivity, newthickness, roadMat,
                            self.road.vegCoverage, self.road.layerTemp[0], self.road.horizontal, self.road._name)

        # Define Rural Element
        ruralMat, newthickness = procMat(self.rural, self.MAXTHICKNESS, self.MINTHICKNESS)

        for i in range(self.nSoil):
            # if soil depth is greater then the thickness of the road
            # we add new slices of soil at max thickness until road is greater or equal

            is_soildepth_equal = self.is_near_zero(self.depth_soil[i][0] - sum(newthickness), 1e-15)

            if is_soildepth_equal or (self.depth_soil[i][0] > sum(newthickness)):
                while self.depth_soil[i][0] > sum(newthickness):
                    newthickness.append(self.MAXTHICKNESS)
                    ruralMat.append(self.SOIL)

                self.soilindex2 = i
                break

        self.rural = Element(self.rural.albedo, self.rural.emissivity, newthickness,
                             ruralMat, self.rural.vegCoverage, self.rural.layerTemp[0], self.rural.horizontal, self.rural._name)

    def hvac_autosize(self):
        """ Section 6 - HVAC Autosizing (unlimited cooling & heating) """

        for i in range(len(self.BEM)):
            if self.is_near_zero(self.autosize) == False:
                self.BEM[i].building.coolCap = 9999.
                self.BEM[i].building.heatCap = 9999.

    def simulate(self):
        """ Section 7 - uwg main section

            self.N                  # Total hours in simulation
            self.ph                 # per hour
            self.dayType            # 3=Sun, 2=Sat, 1=Weekday
            self.ceil_time_step     # simulation timestep (dt) fitted to weather file timestep

            # Output of object instance vector
            self.WeatherData        # Nx1 vector of forc instance
            self.UCMData            # Nx1 vector of UCM instance
            self.UBLData            # Nx1 vector of UBL instance
            self.RSMData            # Nx1 vector of RSM instance
            self.USMData            # Nx1 vector of USM instance
        """

        self.N = int(self.simTime.days * 24)       # total number of hours in simulation
        n = 0                                      # weather time step counter
        self.ph = self.simTime.dt/3600.            # dt (simulation time step) in hours

        # Data dump variables
        time = range(self.N)

        self.WeatherData = [None for x in range(self.N)]
        self.UCMData = [None for x in range(self.N)]
        self.UBLData = [None for x in range(self.N)]
        self.RSMData = [None for x in range(self.N)]
        self.USMData = [None for x in range(self.N)]

        print('\nSimulating new temperature and humidity values for {} days from {}/{}.\n'.format(
            int(self.nDay), int(self.Month), int(self.Day)))
        self.logger.info("Start simulation")

        for it in range(1, self.simTime.nt, 1):  # for every simulation time-step (i.e 5 min) defined by uwg
            # Update water temperature (estimated)
            if self.is_near_zero(self.nSoil):
                # for BUBBLE/CAPITOUL/Singapore only
                self.forc.deepTemp = sum(self.forcIP.temp)/float(len(self.forcIP.temp))
                self.forc.waterTemp = sum(
                    self.forcIP.temp)/float(len(self.forcIP.temp)) - 10.      # for BUBBLE/CAPITOUL/Singapore only
            else:
                # soil temperature by depth, by month
                self.forc.deepTemp = self.Tsoil[self.soilindex1][self.simTime.month-1]
                self.forc.waterTemp = self.Tsoil[2][self.simTime.month-1]

            # There's probably a better way to update the weather...
            self.simTime.UpdateDate()

            self.logger.info("\n{0} m={1}, d={2}, h={3}, s={4}".format(
                __name__, self.simTime.month, self.simTime.day, self.simTime.secDay/3600., self.simTime.secDay))

            # simulation time increment raised to weather time step
            self.ceil_time_step = int(math.ceil(it * self.ph))-1
            # minus one to be consistent with forcIP list index
            # Updating forcing instance
            # horizontal Infrared Radiation Intensity (W m-2)
            self.forc.infra = self.forcIP.infra[self.ceil_time_step]
            # wind speed (m s-1)
            self.forc.wind = max(self.forcIP.wind[self.ceil_time_step], self.geoParam.windMin)
            self.forc.uDir = self.forcIP.uDir[self.ceil_time_step]          # wind direction
            # specific humidty (kg kg-1)
            self.forc.hum = self.forcIP.hum[self.ceil_time_step]
            self.forc.pres = self.forcIP.pres[self.ceil_time_step]          # Pressure (Pa)
            self.forc.temp = self.forcIP.temp[self.ceil_time_step]          # air temperature (C)
            self.forc.rHum = self.forcIP.rHum[self.ceil_time_step]          # Relative humidity (%)
            self.forc.prec = self.forcIP.prec[self.ceil_time_step]          # Precipitation (mm h-1)
            # horizontal solar diffuse radiation (W m-2)
            self.forc.dif = self.forcIP.dif[self.ceil_time_step]
            # normal solar direct radiation (W m-2)
            self.forc.dir = self.forcIP.dir[self.ceil_time_step]
            # Canyon humidity (absolute) same as rural
            self.UCM.canHum = copy.copy(self.forc.hum)

            # Update solar flux
            self.solar = SolarCalcs(self.UCM, self.BEM, self.simTime,
                                    self.RSM, self.forc, self.geoParam, self.rural)
            self.rural, self.UCM, self.BEM = self.solar.solarcalcs()

            # Update building & traffic schedule
            # Assign day type (1 = weekday, 2 = sat, 3 = sun/other)
            if self.is_near_zero(self.simTime.julian % 7):
                self.dayType = 3                                        # Sunday
            elif self.is_near_zero(self.simTime.julian % 7 - 6.):
                self.dayType = 2                                        # Saturday
            else:
                self.dayType = 1                                        # Weekday

            # Update anthropogenic heat load for each hour (building & UCM)
            self.UCM.sensAnthrop = self.sensAnth * (self.SchTraffic[self.dayType-1][self.simTime.hourDay])

            # Update the energy components for building types defined in initialize.uwg
            for i in range(len(self.BEM)):
                # Set temperature
                self.BEM[i].building.coolSetpointDay = self.Sch[i].Cool[self.dayType -
                                                                        1][self.simTime.hourDay] + 273.15  # add from temperature schedule for cooling
                self.BEM[i].building.coolSetpointNight = self.BEM[i].building.coolSetpointDay
                self.BEM[i].building.heatSetpointDay = self.Sch[i].Heat[self.dayType -
                                                                        1][self.simTime.hourDay] + 273.15  # add from temperature schedule for heating
                self.BEM[i].building.heatSetpointNight = self.BEM[i].building.heatSetpointDay

                # Internal Heat Load Schedule (W/m^2 of floor area for Q)
                self.BEM[i].Elec = self.Sch[i].Qelec * self.Sch[i].Elec[self.dayType -
                                                                        1][self.simTime.hourDay]      # Qelec x elec fraction for day
                self.BEM[i].Light = self.Sch[i].Qlight * self.Sch[i].Light[self.dayType -
                                                                           1][self.simTime.hourDay]    # Qlight x light fraction for day
                self.BEM[i].Nocc = self.Sch[i].Nocc * self.Sch[i].Occ[self.dayType -
                                                                      1][self.simTime.hourDay]        # Number of occupants x occ fraction for day
                # Sensible Q occupant * fraction occupant sensible Q * number of occupants
                self.BEM[i].Qocc = self.sensOcc * (1 - self.LatFOcc) * self.BEM[i].Nocc

                # SWH and ventilation schedule
                self.BEM[i].SWH = self.Sch[i].Vswh * self.Sch[i].SWH[self.dayType -
                                                                     1][self.simTime.hourDay]          # litres per hour x SWH fraction for day
                # m^3/s/m^2 of floor
                self.BEM[i].building.vent = self.Sch[i].Vent
                self.BEM[i].Gas = self.Sch[i].Qgas * self.Sch[i].Gas[self.dayType -
                                                                     1][self.simTime.hourDay]          # Gas Equip Schedule, per m^2 of floor

                # This is quite messy, should update
                # Update internal heat and corresponding fractional loads
                intHeat = self.BEM[i].Light + self.BEM[i].Elec + self.BEM[i].Qocc
                # W/m2 from light, electricity, occupants
                self.BEM[i].building.intHeatDay = intHeat
                self.BEM[i].building.intHeatNight = intHeat
                # fraction of radiant heat from light and equipment of whole internal heat
                self.BEM[i].building.intHeatFRad = (
                    self.RadFLight * self.BEM[i].Light + self.RadFEquip * self.BEM[i].Elec) / intHeat
                # fraction of latent heat (from occupants) of whole internal heat
                self.BEM[i].building.intHeatFLat = self.LatFOcc * \
                    self.sensOcc * self.BEM[i].Nocc/intHeat

                # Update envelope temperature layers
                self.BEM[i].T_wallex = self.BEM[i].wall.layerTemp[0]
                self.BEM[i].T_wallin = self.BEM[i].wall.layerTemp[-1]
                self.BEM[i].T_roofex = self.BEM[i].roof.layerTemp[0]
                self.BEM[i].T_roofin = self.BEM[i].roof.layerTemp[-1]

            # Update rural heat fluxes & update vertical diffusion model (VDM)
            self.rural.infra = self.forc.infra - self.rural.emissivity * self.SIGMA * \
                self.rural.layerTemp[0]**4.    # Infrared radiation from rural road

            self.rural.SurfFlux(self.forc, self.geoParam, self.simTime,
                                self.forc.hum, self.forc.temp, self.forc.wind, 2., 0.)
            self.RSM.VDM(self.forc, self.rural, self.geoParam, self.simTime)

            # Calculate urban heat fluxes, update UCM & UBL
            self.UCM, self.UBL, self.BEM = urbflux(
                self.UCM, self.UBL, self.BEM, self.forc, self.geoParam, self.simTime, self.RSM)
            self.UCM.UCModel(self.BEM, self.UBL.ublTemp, self.forc, self.geoParam)
            self.UBL.UBLModel(self.UCM, self.RSM, self.rural,
                              self.forc, self.geoParam, self.simTime)

            """
            # Experimental code to run diffusion model in the urban area
            # N.B Commented out in python uwg because computed wind speed in
            # urban VDM: y = =0.84*ln((2-x/20)/0.51) results in negative log
            # for building heights >= 40m.

            Uroad = copy.copy(self.UCM.road)
            Uroad.sens = copy.copy(self.UCM.sensHeat)
            Uforc = copy.copy(self.forc)
            Uforc.wind = copy.copy(self.UCM.canWind)
            Uforc.temp = copy.copy(self.UCM.canTemp)
            self.USM.VDM(Uforc,Uroad,self.geoParam,self.simTime)
            """

            self.logger.info("dbT = {}".format(self.UCM.canTemp-273.15))
            if n > 0:
                logging.info("dpT = {}".format(self.UCM.Tdp))
                logging.info("RH  = {}".format(self.UCM.canRHum))

            if self.is_near_zero(self.simTime.secDay % self.simTime.timePrint) and n < self.N:

                self.logger.info("{0} ----sim time step = {1}----\n\n".format(__name__, n))

                self.WeatherData[n] = copy.copy(self.forc)
                _Tdb, _w, self.UCM.canRHum, _h, self.UCM.Tdp, _v = psychrometrics(
                    self.UCM.canTemp, self.UCM.canHum, self.forc.pres)

                self.UBLData[n] = copy.copy(self.UBL)
                self.UCMData[n] = copy.copy(self.UCM)
                self.RSMData[n] = copy.copy(self.RSM)

                self.logger.info("dbT = {}".format(self.UCMData[n].canTemp-273.15))
                self.logger.info("dpT = {}".format(self.UCMData[n].Tdp))
                self.logger.info("RH  = {}".format(self.UCMData[n].canRHum))

                n += 1

    def write_epw(self):
        """ Section 8 - Writing new EPW file
        """
        epw_prec = self.epw_precision  # precision of epw file input

        for iJ in range(len(self.UCMData)):
            # [iJ+self.simTime.timeInitial-8] = increments along every weather timestep in epw
            # [6 to 21]                       = column data of epw
            self.epwinput[iJ+self.simTime.timeInitial-8][6] = "{0:.{1}f}".format(
                self.UCMData[iJ].canTemp - 273.15, epw_prec)  # dry bulb temperature  [?C]
            # dew point temperature [?C]
            self.epwinput[iJ+self.simTime.timeInitial -
                          8][7] = "{0:.{1}f}".format(self.UCMData[iJ].Tdp, epw_prec)
            # relative humidity     [%]
            self.epwinput[iJ+self.simTime.timeInitial -
                          8][8] = "{0:.{1}f}".format(self.UCMData[iJ].canRHum, epw_prec)
            self.epwinput[iJ+self.simTime.timeInitial-8][21] = "{0:.{1}f}".format(
                self.WeatherData[iJ].wind, epw_prec)        # wind speed [m/s]

        # Writing new EPW file
        epw_new_id = open(self.newPathName, "w")

        for i in range(8):
            new_epw_line = '{}\n'.format(reduce(lambda x, y: x+","+y, self._header[i]))
            epw_new_id.write(new_epw_line)

        for i in range(len(self.epwinput)):
            printme = ""
            for ei in range(34):
                printme += "{}".format(self.epwinput[i][ei]) + ','
            printme = printme + "{}".format(self.epwinput[i][ei])
            new_epw_line = "{0}\n".format(printme)
            epw_new_id.write(new_epw_line)

        epw_new_id.close()

        print("New climate file '{}' is generated at {}.".format(
            self.destinationFileName, self.destinationDir))

    def run(self):

        # run main class methods
        self.read_epw()
        self.set_input()
        self.init_BEM_obj()
        self.init_input_obj()
        self.hvac_autosize()
        self.simulate()
        self.write_epw()


def procMat(materials, max_thickness, min_thickness):
    """ Processes material layer so that a material with single
    layer thickness is divided into two and material layer that is too
    thick is subdivided
    """
    newmat = []
    newthickness = []
    k = materials.layerThermalCond
    Vhc = materials.layerVolHeat

    if len(materials.layerThickness) > 1:

        for j in range(len(materials.layerThickness)):
            # Break up each layer that's more than max thickness (0.05m)
            if materials.layerThickness[j] > max_thickness:
                nlayers = math.ceil(materials.layerThickness[j]/float(max_thickness))
                for i in range(int(nlayers)):
                    newmat.append(Material(k[j], Vhc[j], name=materials._name))
                    newthickness.append(materials.layerThickness[j]/float(nlayers))
            # Material that's less then min_thickness is not added.
            elif materials.layerThickness[j] < min_thickness:
                print("WARNING: Material '{}' layer found too thin (<{:.2f}cm), ignored.").format(
                    materials._name, min_thickness*100)
            else:
                newmat.append(Material(k[j], Vhc[j], name=materials._name))
                newthickness.append(materials.layerThickness[j])

    else:

        # Divide single layer into two (uwg assumes at least 2 layers)
        if materials.layerThickness[0] > max_thickness:
            nlayers = math.ceil(materials.layerThickness[0]/float(max_thickness))
            for i in range(int(nlayers)):
                newmat.append(Material(k[0], Vhc[0], name=materials._name))
                newthickness.append(materials.layerThickness[0]/float(nlayers))
        # Material should be at least 1cm thick, so if we're here,
        # should give warning and stop. Only warning given for now.
        elif materials.layerThickness[0] < min_thickness*2:
            newthickness = [min_thickness/2., min_thickness/2.]
            newmat = [Material(k[0], Vhc[0], name=materials._name),
                      Material(k[0], Vhc[0], name=materials._name)]
            print("WARNING: a thin (<2cm) single material '{}' layer found. May cause error.".format(
                materials._name))
        else:
            newthickness = [materials.layerThickness[0]/2., materials.layerThickness[0]/2.]
            newmat = [Material(k[0], Vhc[0], name=materials._name),
                      Material(k[0], Vhc[0], name=materials._name)]
    return newmat, newthickness



from __future__ import division


class Forcing (object):
    """
    FORCING
    Force method to estimate deep ground/water temperature

    self.deepTemp                # deep soil temperature (K)
    self.waterTemp               # ground water temp, set to temp at 2m
    self.infra                   # horizontal Infrared Radiation Intensity (W m-2)
    self.uDir                    # wind direction
    self.hum                     # specific humidty (kg kg-1)
    self.pres                    # Pressure (Pa)
    self.temp                    # air temperature (C)
    self.rHum                    # Relative humidity (%)
    self.dir                     # normal solar direct radiation (W m-2)
    self.dif                     # horizontal solar diffuse radiation (W m-2)
                                 # ...Amount of solar radiation received from the sky
                                 # ...(excluding the solar disk) on a horizontal surface
    self.prec                    # Precipitation (mm h-1)
    self.winde                   # wind speed (m s-1)

    """

    def __init__(self,staTemp=None,weather=None):
        # weather: Weather obj
        # staTemp: list of hourly air temperature for simulation period
        # Define default values for instance variables, when the type can be mutable:
        if staTemp==None and weather==None:
            self.deepTemp = None
            self.waterTemp = None
            self.infra = None
            self.uDir = None
            self.hum  = None
            self.pres = None
            self.temp = None
            self.rHum = None
            self.dir = None
            self.dif = None
            self.prec = None
            self.wind = None
        else:
            self.deepTemp = sum(staTemp)/float(len(staTemp))
            self.waterTemp = sum(staTemp)/float(len(staTemp))
            self.infra = weather.staInfra
            self.uDir = weather.staUdir
            self.hum  = weather.staHum
            self.pres = weather.staPres
            self.temp = weather.staTemp
            self.rHum = weather.staRhum
            self.dir = weather.staDir
            self.dif = weather.staDif
            self.prec = [p/3.6e6 for p in weather.staRobs]
            self.wind = weather.staUmod

    def __repr__(self):
        return "Forcing: deepT={a}, waterT={b}".format(
            a=int(self.deepTemp) if self.deepTemp else None,
            b=int(self.waterTemp) if self.waterTemp else None
            )



from __future__ import print_function


class BEMDef(object):
    """
    Building Energy Model (BEM) class definition


    attributes:
        building;       # building class type
        mass;           # mass element class
        wall;           # wall element class
        roof;           # roof element class

        frac;           # fraction of the urban floor space of this typology
        fl_area;        # Total building floor area in the urban area

        ElecTotal;      # Actual total electricity (W/m^2)
        Elec;           # Actual electricity consumption(W/m^2)
        Gas;            # Actual gas consumption(W/m^2)
        Light;          # Actual light (W/m^2)
        Qocc;           # Actual heat load from occupant (W/m^2)
        SWH;            # Actual hot water usage
        Nocc;           # Actual number of people

        T_wallex;       # Wall surface temp (ext)
        T_wallin;       # Wall surface temp (int)
        T_roofex;       # Roof surface temp
        T_roofin;       #
        Q_wallex;       # Wall surface flux (ext)
        Q_wallin;       # Wall surface flux (int)
        Q_roofex;
        Q_roofin;
    """

    def __init__(self,building,mass,wall,roof,frac):
        self.building = building    # building class type
        self.mass = mass            # mass element class
        self.wall = wall            # wall element class
        self.roof = roof            # roof element class
        self.frac = frac            # fraction of the urban floor space of this typology (~ bld_footprint/urba_footprint ?)

    def __repr__(self):
        return "BEMDef: Type = {}, Zone = {}, Era = {}, Construction = {}".format(
            self.building.Type,
            self.building.Zone,
            self.building.Era,
            self.wall._name
            )
